import freeOrionAIInterface as fo
import math
import random
from collections.abc import Iterable, Sequence
from logging import debug, error, warning
from operator import itemgetter
from typing import (
    NamedTuple,
    NewType,
    Union,
)

import AIDependencies
import AIstate
import FleetUtilsAI
import MilitaryAI
import PlanetUtilsAI
import PriorityAI
from AIDependencies import INVALID_ID, Tags
from aistate_interface import get_aistate
from buildings import BuildingType, BuildingTypeBase, Shipyard, get_empire_drydocks
from character.character_module import Aggression
from colonization import rate_planetary_piloting
from colonization.colony_score import MINIMUM_COLONY_SCORE
from colonization.rate_pilots import GREAT_PILOT_RATING
from common.fo_typing import BuildingName, PlanetId, SystemId
from empire.buildings_locations import get_best_pilot_facilities
from empire.colony_builders import (
    can_build_colony_for_species,
    can_build_only_sly_colonies,
    get_colony_builder_locations,
    get_colony_builders,
)
from empire.pilot_rating import (
    best_pilot_rating,
    get_pilot_ratings,
    get_rating_for_planet,
    medium_pilot_rating,
)
from empire.ship_builders import (
    can_build_ship_for_species,
    get_ship_builder_locations,
    has_shipyard,
)
from EnumsAI import (
    EmpireProductionTypes,
    FocusType,
    MissionType,
    PriorityType,
    ShipRoleType,
    get_priority_production_types,
)
from expansion_plans import get_colonisable_planet_ids
from freeorion_tools import get_named_real, ppstring, tech_is_complete
from production import print_building_list, print_capital_info, print_production_queue
from turn_state import (
    get_all_empire_planets,
    get_colonized_planets,
    get_colonized_planets_in_system,
    get_empire_outposts,
    get_empire_planets_by_species,
    get_empire_planets_with_species,
    get_inhabited_planets,
    get_owned_planets,
    get_owned_planets_in_system,
    population_with_industry_focus,
)
from turn_state.design import get_best_ship_info, get_best_ship_ratings
from universe.system_network import get_neighbors


def get_priority_locations() -> frozenset[PlanetId]:
    priority_facilities = [
        "BLD_SHIPYARD_ENRG_SOLAR",
        "BLD_SHIPYARD_CON_GEOINT",
        "BLD_SHIPYARD_AST_REF",
        Shipyard.ENRG_COMP.value,
    ]
    # TODO: also cover good troopship locations
    return frozenset(loc for building in priority_facilities for loc in get_best_pilot_facilities(building))


# set by ResourceAI
candidate_for_translator = None


def translators_wanted() -> int:
    """
    How many near universal translators we'd like to build.
    Function is also used by ResearchAI to determine when to research the prerequisites.
    """
    pid = PlanetUtilsAI.get_capital()
    # planet is needed to determine the cost. Without a capital we have bigger problems anyway...
    # At the beginning of the game influence_priority fluctuates strongly and that early in the game there are
    # surely more important things to build.
    if pid == INVALID_ID or fo.currentTurn() < 30:
        return 0.0
    building_type = BuildingType.TRANSLATOR
    translator_cost = building_type.production_cost(pid)
    influence_priority = get_aistate().get_priority(PriorityType.RESOURCE_INFLUENCE)
    num_species = len(get_empire_planets_by_species())
    num_enqueued = len(building_type.queued_at())
    num_built = len(building_type.built_at())
    # first one gives a policy slot
    first_bonus = 30 if num_enqueued + num_built == 0 else 0
    importance = 6 * (influence_priority + first_bonus) / translator_cost * num_species**0.25 - num_enqueued
    debug(
        f"translators_wanted: influence_priority: {influence_priority}, translator_cost: {translator_cost}, "
        f"built: {num_built}, enqueued: {num_enqueued}, num_species: {num_species}, turn: {fo.currentTurn()}, "
        f"importance: {importance}"
    )
    return int(importance)


def _get_capital_info() -> tuple[PlanetId, "fo.planet", SystemId]:
    capital_id = PlanetUtilsAI.get_capital()
    if capital_id is None or capital_id == INVALID_ID:
        homeworld = None
        capital_system_id = None
    else:
        homeworld = fo.getUniverse().getPlanet(capital_id)
        capital_system_id = homeworld.systemID
    return capital_id, homeworld, capital_system_id


def _first_turn_first_time():
    """
    Return true if generate production order at the first time.

    If game is loaded on the first turn this code should return false.
    """
    return fo.currentTurn() == 1 and len(AIstate.opponentPlanetIDs) == 0 and len(fo.getEmpire().productionQueue) == 0


def _first_turn_action():
    if not _first_turn_first_time():
        return

    init_build_nums = [(PriorityType.PRODUCTION_EXPLORATION, 2)]
    if can_build_only_sly_colonies():
        init_build_nums.append((PriorityType.PRODUCTION_COLONISATION, 1))
    else:
        init_build_nums.append((PriorityType.PRODUCTION_OUTPOST, 1))
    for ship_type, num_ships in init_build_nums:
        best_design_id, _, build_choices = get_best_ship_info(ship_type)
        if best_design_id is not None:
            for _ in range(num_ships):
                fo.issueEnqueueShipProductionOrder(best_design_id, build_choices[0])
            fo.updateProductionQueue()


def get_building_allocations() -> float:
    empire = fo.getEmpire()
    return sum(e.allocation for e in empire.productionQueue if e.buildType == EmpireProductionTypes.BT_BUILDING)


# TODO Move Building names to AIDependencies to avoid typos and for IDE-Support
def generate_production_orders():  # noqa: C901
    """Generate production orders."""
    # first check ship designs
    # next check for buildings etc, that could be placed on queue regardless of locally available PP
    # next loop over resource groups, adding buildings & ships
    print_production_queue()
    universe = fo.getUniverse()

    debug("Production Queue Management:")
    empire = fo.getEmpire()
    debug("")
    debug("  Total Available Production Points: %s" % empire.productionPoints)
    print_building_list()

    aistate = get_aistate()

    _first_turn_action()
    building_expense = 0.0
    building_ratio = aistate.character.preferred_building_ratio([0.4, 0.35, 0.30])
    capital_id, homeworld, capital_system_id = _get_capital_info()
    current_turn = fo.currentTurn()
    building_expense += get_building_allocations()
    if not homeworld:
        debug("if no capital, no place to build, should get around to capturing or colonizing a new one")  # TODO
    else:
        debug("Empire priority_id %d has current Capital %s:" % (empire.empireID, homeworld.name))
        print_capital_info(homeworld)
        capital_buildings = [universe.getBuilding(bldg).buildingTypeName for bldg in homeworld.buildingIDs]

        possible_building_type_ids = []
        for type_id in empire.availableBuildingTypes:
            # Apparently, when loading a saved game from another version, availableBuildingTypes may return
            # buildings that are not in our script.
            fo_building_type = fo.getBuildingType(type_id)
            if fo_building_type and fo_building_type.canBeProduced(empire.empireID, homeworld.id):
                possible_building_type_ids.append(type_id)

        if possible_building_type_ids:
            debug("Possible building types to build:")
            for type_id in possible_building_type_ids:
                building_type = fo.getBuildingType(type_id)
                debug(
                    "    {} cost: {}  time: {}".format(
                        building_type.name,
                        building_type.productionCost(empire.empireID, homeworld.id),
                        building_type.productionTime(empire.empireID, homeworld.id),
                    )
                )

            possible_building_types = [fo.getBuildingType(type_id).name for type_id in possible_building_type_ids]

            queued_building_names = _get_queued_buildings(capital_id)

            if "BLD_AUTO_HISTORY_ANALYSER" in possible_building_types:
                for pid in find_automatic_historic_analyzer_candidates():
                    res = fo.issueEnqueueBuildingProductionOrder("BLD_AUTO_HISTORY_ANALYSER", pid)
                    debug(
                        "Enqueueing BLD_AUTO_HISTORY_ANALYSER at planet %s - result %d" % (universe.getPlanet(pid), res)
                    )
                    if res:
                        cost, time = empire.productionCostAndTime(
                            empire.productionQueue[empire.productionQueue.size - 1]
                        )
                        building_expense += cost / time
                        res = fo.issueRequeueProductionOrder(empire.productionQueue.size - 1, 0)  # move to front
                        debug(
                            "Requeueing %s to front of build queue, with result %d" % ("BLD_AUTO_HISTORY_ANALYSER", res)
                        )

            # TODO: check existence of BLD_INDUSTRY_CENTER (and other buildings) in other locations in case we captured it
            if (
                (empire.productionPoints > 40 or ((current_turn > 40) and (population_with_industry_focus() >= 20)))
                and ("BLD_INDUSTRY_CENTER" in possible_building_types)
                and ("BLD_INDUSTRY_CENTER" not in (capital_buildings + queued_building_names))
                and (building_expense < building_ratio * empire.productionPoints)
            ):
                res = fo.issueEnqueueBuildingProductionOrder("BLD_INDUSTRY_CENTER", homeworld.id)
                debug("Enqueueing BLD_INDUSTRY_CENTER, with result %d" % res)
                if res:
                    cost, time = empire.productionCostAndTime(empire.productionQueue[empire.productionQueue.size - 1])
                    building_expense += cost / time

            if ("BLD_SHIPYARD_BASE" in possible_building_types) and (
                "BLD_SHIPYARD_BASE" not in (capital_buildings + queued_building_names)
            ):
                try:
                    res = fo.issueEnqueueBuildingProductionOrder("BLD_SHIPYARD_BASE", homeworld.id)
                    debug("Enqueueing BLD_SHIPYARD_BASE, with result %d" % res)
                except:  # noqa: E722
                    warning("Can't build shipyard at new capital, probably no population; we're hosed")

            for building_name in ["BLD_SHIPYARD_ORG_ORB_INC"]:
                if (
                    (building_name in possible_building_types)
                    and (building_name not in (capital_buildings + queued_building_names))
                    and (building_expense < building_ratio * empire.productionPoints)
                ):
                    try:
                        res = fo.issueEnqueueBuildingProductionOrder(building_name, homeworld.id)
                        debug("Enqueueing %s at capital, with result %d" % (building_name, res))
                        if res:
                            cost, time = empire.productionCostAndTime(
                                empire.productionQueue[empire.productionQueue.size - 1]
                            )
                            building_expense += cost / time
                            res = fo.issueRequeueProductionOrder(empire.productionQueue.size - 1, 0)  # move to front
                            debug("Requeueing %s to front of build queue, with result %d" % (building_name, res))
                    except:  # noqa: E722
                        error("Exception triggered and caught: ", exc_info=True)

            building_type = BuildingType.PALACE
            if building_type.available() and not building_type.built_or_queued_at():
                building_expense += _try_enqueue(building_type, capital_id, at_front=True, ignore_dislike=True)

            # ok, BLD_NEUTRONIUM_SYNTH is not currently unlockable, but just in case... ;-p
            if ("BLD_NEUTRONIUM_SYNTH" in possible_building_types) and (
                "BLD_NEUTRONIUM_SYNTH" not in (capital_buildings + queued_building_names)
            ):
                res = fo.issueEnqueueBuildingProductionOrder("BLD_NEUTRONIUM_SYNTH", homeworld.id)
                debug("Enqueueing BLD_NEUTRONIUM_SYNTH, with result %d" % res)
                if res:
                    res = fo.issueRequeueProductionOrder(empire.productionQueue.size - 1, 0)  # move to front
                    debug("Requeueing BLD_NEUTRONIUM_SYNTH to front of build queue, with result %d" % res)

    max_defense_portion = aistate.character.max_defense_portion()
    if aistate.character.check_orbital_production():
        sys_orbital_defenses = {}
        queued_defenses = {}
        defense_allocation = 0.0
        target_orbitals = aistate.character.target_number_of_orbitals()
        debug("Orbital Defense Check -- target Defense Orbitals: %s" % target_orbitals)
        for element in empire.productionQueue:
            if (element.buildType == EmpireProductionTypes.BT_SHIP) and (
                aistate.get_ship_role(element.designID) == ShipRoleType.BASE_DEFENSE
            ):
                planet = universe.getPlanet(element.locationID)
                if not planet:
                    error("Problem getting Planet for build loc %s" % element.locationID)
                    continue
                sys_id = planet.systemID
                queued_defenses[sys_id] = queued_defenses.get(sys_id, 0) + element.blocksize * element.remaining
                defense_allocation += element.allocation
        debug(
            "Queued Defenses: %s",
            ppstring([(str(universe.getSystem(sid)), num) for sid, num in queued_defenses.items()]),
        )
        for sys_id, pids in get_colonized_planets().items():
            if aistate.systemStatus.get(sys_id, {}).get("fleetThreat", 1) > 0:
                continue  # don't build orbital shields if enemy fleet present
            if defense_allocation > max_defense_portion * empire.productionPoints:
                break
            sys_orbital_defenses[sys_id] = 0
            fleets_here = aistate.systemStatus.get(sys_id, {}).get("myfleets", [])
            for fid in fleets_here:
                if aistate.get_fleet_role(fid) == MissionType.ORBITAL_DEFENSE:
                    debug(
                        "Found %d existing Orbital Defenses in %s :"
                        % (aistate.fleetStatus.get(fid, {}).get("nships", 0), universe.getSystem(sys_id))
                    )
                    sys_orbital_defenses[sys_id] += aistate.fleetStatus.get(fid, {}).get("nships", 0)
            for pid in pids:
                sys_orbital_defenses[sys_id] += queued_defenses.get(pid, 0)
            if sys_orbital_defenses[sys_id] < target_orbitals:
                num_needed = target_orbitals - sys_orbital_defenses[sys_id]
                for pid in pids:
                    best_design_id, col_design, build_choices = get_best_ship_info(
                        PriorityType.PRODUCTION_ORBITAL_DEFENSE, pid
                    )
                    if not best_design_id:
                        debug("no orbital defenses can be built at %s", PlanetUtilsAI.planet_string(pid))
                        continue
                    retval = fo.issueEnqueueShipProductionOrder(best_design_id, pid)
                    debug("queueing %d Orbital Defenses at %s" % (num_needed, PlanetUtilsAI.planet_string(pid)))
                    if retval != 0:
                        if num_needed > 1:
                            fo.issueChangeProductionQuantityOrder(empire.productionQueue.size - 1, 1, num_needed)
                        cost, time = empire.productionCostAndTime(
                            empire.productionQueue[empire.productionQueue.size - 1]
                        )
                        defense_allocation += (
                            empire.productionQueue[empire.productionQueue.size - 1].blocksize * cost / time
                        )
                        fo.issueRequeueProductionOrder(empire.productionQueue.size - 1, 0)  # move to front
                        break

    info = _build_basic_shipyards()
    queued_shipyard_pids = info.queued_shipyard_pids
    colony_systems = info.colony_systems
    top_pilot_systems = info.top_pilot_systems
    blackhole_pilots, red_pilots, building_expense = _build_energy_shipyards(
        queued_shipyard_pids, colony_systems, building_ratio, building_expense
    )
    _build_ship_facilities(Shipyard.ORG_ORB_INC)
    # gating by life cycle manipulation helps delay these until they are closer to being worthwhile
    if tech_is_complete(AIDependencies.GRO_LIFE_CYCLE) or empire.researchProgress(AIDependencies.GRO_LIFE_CYCLE) > 0:
        for building_type in (Shipyard.XENO_FACILITY, Shipyard.ORG_CELL_GRO_CHAMB):
            _build_ship_facilities(building_type)
    building_expense += _build_asteroid_processor(top_pilot_systems, queued_shipyard_pids)

    building_expense += _build_gas_giant_generator()
    building_expense += _build_translator()
    building_expense += _build_regional_administration()
    building_expense += _build_military_command()

    building_name = "BLD_SOL_ORB_GEN"
    if empire.buildingTypeAvailable(building_name) and aistate.character.may_build_building(building_name):
        already_got_one = 99
        for pid in get_all_empire_planets():
            planet = universe.getPlanet(pid)
            if planet and building_name in [
                bld.buildingTypeName for bld in map(universe.getBuilding, planet.buildingIDs)
            ]:
                system = universe.getSystem(planet.systemID)
                if system and system.starType < already_got_one:
                    already_got_one = system.starType
        best_type = fo.starType.white
        best_locs = AIstate.empireStars.get(fo.starType.blue, []) + AIstate.empireStars.get(fo.starType.white, [])
        if not best_locs:
            best_type = fo.starType.orange
            best_locs = AIstate.empireStars.get(fo.starType.yellow, []) + AIstate.empireStars.get(
                fo.starType.orange, []
            )
        if (not best_locs) or (already_got_one < 99 and already_got_one <= best_type):
            pass  # could consider building at a red star if have a lot of PP but somehow no better stars
        else:
            use_new_loc = True
            queued_building_locs = [
                element.locationID for element in (empire.productionQueue) if (element.name == building_name)
            ]
            if queued_building_locs:
                queued_star_types = {}
                for loc in queued_building_locs:
                    planet = universe.getPlanet(loc)
                    if not planet:
                        continue
                    system = universe.getSystem(planet.systemID)
                    queued_star_types.setdefault(system.starType, []).append(loc)
                if queued_star_types:
                    best_queued = sorted(queued_star_types.keys())[0]
                    if best_queued > best_type:  # i.e., best_queued is yellow, best_type available is blue or white
                        pass  # should probably evaluate cancelling the existing one under construction
                    else:
                        use_new_loc = False
            if use_new_loc:  # (of course, may be only loc, not really new)
                if not homeworld:
                    use_sys = best_locs[0]  # as good as any
                else:
                    distance_map = {}
                    for sys_id in best_locs:  # want to build close to capital for defense
                        if sys_id == INVALID_ID:
                            continue
                        try:
                            distance_map[sys_id] = universe.jumpDistance(homeworld.systemID, sys_id)
                        except:  # noqa: E722
                            pass
                    use_sys = ([(-1, INVALID_ID)] + sorted([(dist, sys_id) for sys_id, dist in distance_map.items()]))[
                        :2
                    ][-1][
                        -1
                    ]  # kinda messy, but ensures a value
                if use_sys != INVALID_ID:
                    try:
                        use_loc = get_owned_planets_in_system(use_sys)[0]
                        res = fo.issueEnqueueBuildingProductionOrder(building_name, use_loc)
                        debug(
                            "Enqueueing %s at planet %d (%s) , with result %d",
                            building_name,
                            use_loc,
                            universe.getPlanet(use_loc).name,
                            res,
                        )
                        if res:
                            cost, time = empire.productionCostAndTime(
                                empire.productionQueue[empire.productionQueue.size - 1]
                            )
                            building_expense += cost / time  # production_queue[production_queue.size -1].blocksize *
                            res = fo.issueRequeueProductionOrder(empire.productionQueue.size - 1, 0)  # move to front
                            debug("Requeueing %s to front of build queue, with result %d", building_name, res)
                    except:  # noqa: E722
                        debug("problem queueing BLD_SOL_ORB_GEN at planet %s of system", use_loc, use_sys)

    building_name = "BLD_ART_BLACK_HOLE"
    if (
        empire.buildingTypeAvailable(building_name)
        and aistate.character.may_build_building(building_name)
        and len(AIstate.empireStars.get(fo.starType.red, [])) > 0
    ):
        already_got_one = False
        for pid in get_all_empire_planets():
            planet = universe.getPlanet(pid)
            if planet and building_name in [
                bld.buildingTypeName for bld in map(universe.getBuilding, planet.buildingIDs)
            ]:
                already_got_one = True  # has been built, needs one turn to activate
        queued_building_locs = [
            element.locationID for element in (empire.productionQueue) if (element.name == building_name)
        ]  # TODO: check that queued locs or already built one are at red stars
        if not blackhole_pilots and len(queued_building_locs) == 0 and (red_pilots or not already_got_one):
            use_loc = None
            nominal_home = homeworld or universe.getPlanet(
                (red_pilots + get_owned_planets_in_system(AIstate.empireStars[fo.starType.red][0]))[0]
            )
            distance_map = {}
            for sys_id in AIstate.empireStars.get(fo.starType.red, []):
                if sys_id == INVALID_ID:
                    continue
                try:
                    distance_map[sys_id] = universe.jumpDistance(nominal_home.systemID, sys_id)
                except:  # noqa: E722
                    pass
            red_sys_list = sorted([(dist, sys_id) for sys_id, dist in distance_map.items()])
            for dist, sys_id in red_sys_list:
                for loc in get_owned_planets_in_system(sys_id):
                    planet = universe.getPlanet(loc)
                    if planet and planet.speciesName not in ["", None]:
                        species = fo.getSpecies(planet.speciesName)
                        if species and "PHOTOTROPHIC" in list(species.tags):
                            break
                else:
                    use_loc = list(
                        set(red_pilots).intersection(get_owned_planets_in_system(sys_id))
                        or get_owned_planets_in_system(sys_id)
                    )[0]
                if use_loc is not None:
                    break
            if use_loc is not None:
                planet_used = universe.getPlanet(use_loc)
                try:
                    res = fo.issueEnqueueBuildingProductionOrder(building_name, use_loc)
                    debug("Enqueueing %s at planet %s , with result %d", building_name, planet_used, res)
                    if res:
                        res = fo.issueRequeueProductionOrder(empire.productionQueue.size - 1, 0)  # move to front
                        debug("Requeueing %s to front of build queue, with result %d", building_name, res)
                except:  # noqa: E722
                    debug(f"problem queueing {building_name} at planet {planet_used}")

    building_name = "BLD_BLACK_HOLE_POW_GEN"
    if empire.buildingTypeAvailable(building_name) and aistate.character.may_build_building(building_name):
        already_got_one = False
        for pid in get_all_empire_planets():
            planet = universe.getPlanet(pid)
            if planet and building_name in [
                bld.buildingTypeName for bld in map(universe.getBuilding, planet.buildingIDs)
            ]:
                already_got_one = True
        queued_building_locs = [
            element.locationID for element in (empire.productionQueue) if (element.name == building_name)
        ]
        if (
            (len(AIstate.empireStars.get(fo.starType.blackHole, [])) > 0)
            and len(queued_building_locs) == 0
            and not already_got_one
        ):
            if not homeworld:
                use_sys = AIstate.empireStars.get(fo.starType.blackHole, [])[0]
            else:
                distance_map = {}
                for sys_id in AIstate.empireStars.get(fo.starType.blackHole, []):
                    if sys_id == INVALID_ID:
                        continue
                    try:
                        distance_map[sys_id] = universe.jumpDistance(homeworld.systemID, sys_id)
                    except:  # noqa: E722
                        pass
                use_sys = ([(-1, INVALID_ID)] + sorted([(dist, sys_id) for sys_id, dist in distance_map.items()]))[:2][
                    -1
                ][
                    -1
                ]  # kinda messy, but ensures a value
            if use_sys != INVALID_ID:
                try:
                    use_loc = get_owned_planets_in_system(use_sys)[0]
                    res = fo.issueEnqueueBuildingProductionOrder(building_name, use_loc)
                    debug(
                        "Enqueueing %s at planet %d (%s) , with result %d",
                        building_name,
                        use_loc,
                        universe.getPlanet(use_loc).name,
                        res,
                    )
                    if res:
                        res = fo.issueRequeueProductionOrder(empire.productionQueue.size - 1, 0)  # move to front
                        debug("Requeueing %s to front of build queue, with result %d" % (building_name, res))
                except:  # noqa: E722
                    warning("problem queueing BLD_BLACK_HOLE_POW_GEN at planet %s of system %s", use_loc, use_sys)

    building_name = "BLD_ENCLAVE_VOID"
    if empire.buildingTypeAvailable(building_name):
        already_got_one = False
        for pid in get_all_empire_planets():
            planet = universe.getPlanet(pid)
            if planet and building_name in [
                bld.buildingTypeName for bld in map(universe.getBuilding, planet.buildingIDs)
            ]:
                already_got_one = True
        queued_locs = [element.locationID for element in (empire.productionQueue) if (element.name == building_name)]
        if len(queued_locs) == 0 and homeworld and not already_got_one:
            try:
                res = fo.issueEnqueueBuildingProductionOrder(building_name, capital_id)
                debug(
                    "Enqueueing %s at planet %d (%s) , with result %d",
                    building_name,
                    capital_id,
                    universe.getPlanet(capital_id).name,
                    res,
                )
                if res:
                    res = fo.issueRequeueProductionOrder(empire.productionQueue.size - 1, 0)  # move to front
                    debug("Requeueing %s to front of build queue, with result %d", building_name, res)
            except:  # noqa: E722
                pass

    building_name = "BLD_GENOME_BANK"
    if empire.buildingTypeAvailable(building_name):
        already_got_one = False
        for pid in get_all_empire_planets():
            planet = universe.getPlanet(pid)
            if planet and building_name in [
                bld.buildingTypeName for bld in map(universe.getBuilding, planet.buildingIDs)
            ]:
                already_got_one = True
        queued_locs = [element.locationID for element in (empire.productionQueue) if (element.name == building_name)]
        if len(queued_locs) == 0 and homeworld and not already_got_one:
            try:
                res = fo.issueEnqueueBuildingProductionOrder(building_name, capital_id)
                debug(
                    "Enqueueing %s at planet %d (%s) , with result %d",
                    building_name,
                    capital_id,
                    universe.getPlanet(capital_id).name,
                    res,
                )
                if res:
                    res = fo.issueRequeueProductionOrder(empire.productionQueue.size - 1, 0)  # move to front
                    debug("Requeueing %s to front of build queue, with result %d", building_name, res)
            except:  # noqa: E722
                pass

    building_name = "BLD_NEUTRONIUM_EXTRACTOR"
    already_got_extractor = False
    if (
        empire.buildingTypeAvailable(building_name)
        and [element.locationID for element in (empire.productionQueue) if (element.name == building_name)] == []
        and AIstate.empireStars.get(fo.starType.neutron, [])
    ):
        # building_type = fo.getBuildingType(building_name)
        for pid in get_all_empire_planets():
            planet = universe.getPlanet(pid)
            if planet:
                building_names = [bld.buildingTypeName for bld in map(universe.getBuilding, planet.buildingIDs)]
                if (
                    planet.systemID in AIstate.empireStars.get(fo.starType.neutron, [])
                    and building_name in building_names
                ) or "BLD_NEUTRONIUM_SYNTH" in building_names:
                    already_got_extractor = True
        if not already_got_extractor:
            if not homeworld:
                use_sys = AIstate.empireStars.get(fo.starType.neutron, [])[0]
            else:
                distance_map = {}
                for sys_id in AIstate.empireStars.get(fo.starType.neutron, []):
                    if sys_id == INVALID_ID:
                        continue
                    try:
                        distance_map[sys_id] = universe.jumpDistance(homeworld.systemID, sys_id)
                    except Exception:
                        warning("Could not get jump distance from %d to %d", homeworld.systemID, sys_id, exc_info=True)
                debug([INVALID_ID] + sorted([(dist, sys_id) for sys_id, dist in distance_map.items()]))
                use_sys = ([(-1, INVALID_ID)] + sorted([(dist, sys_id) for sys_id, dist in distance_map.items()]))[:2][
                    -1
                ][
                    -1
                ]  # kinda messy, but ensures a value
            if use_sys != INVALID_ID:
                try:
                    use_loc = get_owned_planets_in_system(use_sys)[0]
                    res = fo.issueEnqueueBuildingProductionOrder(building_name, use_loc)
                    debug(
                        "Enqueueing %s at planet %d (%s) , with result %d",
                        building_name,
                        use_loc,
                        universe.getPlanet(use_loc).name,
                        res,
                    )
                    if res:
                        res = fo.issueRequeueProductionOrder(empire.productionQueue.size - 1, 0)  # move to front
                        debug("Requeueing %s to front of build queue, with result %d", building_name, res)
                except:  # noqa: E722
                    warning(f"problem queueing BLD_NEUTRONIUM_EXTRACTOR at planet {use_loc} of system {use_sys}")

    _build_ship_facilities(Shipyard.GEO)

    # with current stats the AI considers Titanic Hull superior to Scattered Asteroid, so don't bother building for now
    # TODO: uncomment once dynamic assessment of prospective designs is enabled & indicates building is worthwhile
    _build_ship_facilities(Shipyard.ASTEROID_REF)

    _build_ship_facilities(Shipyard.NEUTRONIUM_FORGE, get_priority_locations())

    colony_ship_map = {}
    for fid in FleetUtilsAI.get_empire_fleet_ids_by_role(MissionType.COLONISATION):
        fleet = universe.getFleet(fid)
        if not fleet:
            continue
        for shipID in fleet.shipIDs:
            ship = universe.getShip(shipID)
            if ship and (aistate.get_ship_role(ship.design.id) == ShipRoleType.CIVILIAN_COLONISATION):
                colony_ship_map.setdefault(ship.speciesName, []).append(1)

    building_name = "BLD_CONC_CAMP"
    building_type = fo.getBuildingType(building_name)
    for pid in get_inhabited_planets():
        planet = universe.getPlanet(pid)
        if not planet:
            continue
        can_build_camp = building_type.canBeProduced(empire.empireID, pid) and empire.buildingTypeAvailable(
            building_name
        )
        t_pop = planet.initialMeterValue(fo.meterType.targetPopulation)
        c_pop = planet.initialMeterValue(fo.meterType.population)
        t_ind = planet.currentMeterValue(fo.meterType.targetIndustry)
        c_ind = planet.currentMeterValue(fo.meterType.industry)
        pop_disqualified = (c_pop <= 32) or (c_pop < 0.9 * t_pop)
        this_spec = planet.speciesName
        safety_margin_met = (
            can_build_colony_for_species(this_spec)
            and (len(get_empire_planets_with_species(this_spec)) + len(colony_ship_map.get(this_spec, [])) >= 2)
        ) or (c_pop >= 50)
        if (
            pop_disqualified or not safety_margin_met
        ):  # check even if not aggressive, etc, just in case acquired planet with a ConcCamp on it
            for bldg in planet.buildingIDs:
                if universe.getBuilding(bldg).buildingTypeName == building_name:
                    res = fo.issueScrapOrder(bldg)
                    debug("Tried scrapping %s at planet %s, got result %d", building_name, planet.name, res)
        elif aistate.character.may_build_building(building_name) and can_build_camp and (t_pop >= 36):
            if (
                (planet.focus == FocusType.FOCUS_GROWTH)
                or (AIDependencies.COMPUTRONIUM_SPECIAL in planet.specials)
                or (pid == capital_id)
            ):
                continue
                # now that focus setting takes these into account, probably works ok to have conc camp, but let's not push it
            queued_building_locs = [
                element.locationID for element in (empire.productionQueue) if (element.name == building_name)
            ]
            if c_pop >= 0.95 * t_pop:
                if pid not in queued_building_locs:
                    if planet.focus in [FocusType.FOCUS_INDUSTRY]:
                        if c_ind >= t_ind + c_pop:
                            continue
                    else:
                        old_focus = planet.focus
                        fo.issueChangeFocusOrder(pid, FocusType.FOCUS_INDUSTRY)
                        universe.updateMeterEstimates([pid])
                        t_ind = planet.currentMeterValue(fo.meterType.targetIndustry)
                        if c_ind >= t_ind + c_pop:
                            fo.issueChangeFocusOrder(pid, old_focus)
                            universe.updateMeterEstimates([pid])
                            continue
                    res = fo.issueEnqueueBuildingProductionOrder(building_name, pid)
                    debug(
                        "Enqueueing %s at planet %d (%s) , with result %d",
                        building_name,
                        pid,
                        universe.getPlanet(pid).name,
                        res,
                    )
                    if res:
                        queued_building_locs.append(pid)
                        fo.issueRequeueProductionOrder(empire.productionQueue.size - 1, 0)  # move to front
                    else:
                        # TODO: enable location condition reporting a la mapwnd BuildDesignatorWnd
                        warning(
                            f"Enqueing Conc Camp at {planet} despite building_type.canBeProduced(empire.empireID, pid) reporting {can_build_camp}"
                        )
    building_expense += _build_scanning_facility()

    _build_orbital_drydock(top_pilot_systems)

    building_name = "BLD_XENORESURRECTION_LAB"
    queued_xeno_lab_locs = [element.locationID for element in (empire.productionQueue) if element.name == building_name]
    for pid in get_all_empire_planets():
        if pid in queued_xeno_lab_locs or not empire.canBuild(fo.buildType.BT_BUILDING, building_name, pid):
            continue
        res = fo.issueEnqueueBuildingProductionOrder(building_name, pid)
        debug("Enqueueing %s at planet %d (%s) , with result %d", building_name, pid, planet.name, res)
        if res:
            res = fo.issueRequeueProductionOrder(empire.productionQueue.size - 1, 2)  # move to near front
            debug("Requeueing %s to front of build queue, with result %d", building_name, res)
            break
        else:
            warning("Failed enqueueing %s at planet %s, got result %d", building_name, planet, res)

    # ignore acquired-under-construction colony buildings for which our empire lacks the species
    queued_clny_bld_locs = [
        element.locationID
        for element in (empire.productionQueue)
        if (element.name.startswith("BLD_COL_") and empire_has_colony_bld_species(element.name))
    ]
    colony_bldg_entries = [
        entry
        for entry in get_colonisable_planet_ids().items()
        if entry[1][0] > MINIMUM_COLONY_SCORE
        and entry[0] not in queued_clny_bld_locs
        and entry[0] in get_empire_outposts()
        and not already_has_completed_colony_building(entry[0])
    ]
    colony_bldg_entries = colony_bldg_entries[: PriorityAI.allottedColonyTargets + 2]
    for entry in colony_bldg_entries:
        pid = entry[0]
        building_name = "BLD_COL_" + entry[1][1][3:]
        planet = universe.getPlanet(pid)
        building_type = fo.getBuildingType(building_name)
        # We may have conquered a planet with a queued colony.
        # If we want to build another species, we have to remove the queued one.
        _remove_other_colonies(pid, building_name)
        if not (building_type and building_type.canBeEnqueued(empire.empireID, pid)):
            continue
        res = fo.issueEnqueueBuildingProductionOrder(building_name, pid)
        debug("Enqueueing %s at planet %d (%s) , with result %d", building_name, pid, planet.name, res)
        if res:
            res = fo.issueRequeueProductionOrder(empire.productionQueue.size - 1, 2)  # move to near front
            debug("Requeueing %s to front of build queue, with result %d", building_name, res)
            break
        else:
            warning("Failed enqueueing %s at planet %s, got result %d" % (building_name, planet, res))

    buildings_to_scrap = ("BLD_EVACUATION", "BLD_GATEWAY_VOID")
    for pid in get_inhabited_planets():
        planet = universe.getPlanet(pid)
        if not planet:
            continue
        for bldg in planet.buildingIDs:
            building_name = universe.getBuilding(bldg).buildingTypeName
            if building_name in buildings_to_scrap:
                res = fo.issueScrapOrder(bldg)
                debug("Tried scrapping %s at planet %s, got result %d", building_name, planet.name, res)

    total_pp_spent = fo.getEmpire().productionQueue.totalSpent
    debug("  Total Production Points Spent: %s", total_pp_spent)

    wasted_pp = max(0, empire.productionPoints - total_pp_spent)
    debug("  Wasted Production Points: %s", wasted_pp)  # TODO: add resource group analysis
    avail_pp = empire.productionPoints - total_pp_spent - 0.0001

    production_queue = empire.productionQueue
    queued_colony_ships = {}
    queued_outpost_ships = 0
    queued_troop_ships = 0

    # TODO: blocked items might not need dequeuing, but rather for supply lines to be un-blockaded
    dequeue_list = []
    fo.updateProductionQueue()
    can_prioritize_troops = False
    for queue_index in range(len(production_queue)):
        element = production_queue[queue_index]
        block_str = (
            "%d x " % element.blocksize
        )  # ["a single ", "in blocks of %d "%element.blocksize][element.blocksize>1]
        debug(
            "    %s%s  requiring %s  more turns; alloc: %.2f PP with cum. progress of %.1f  being built at %s",
            block_str,
            element.name,
            element.turnsLeft,
            element.allocation,
            element.progress,
            universe.getPlanet(element.locationID).name,
        )
        if element.turnsLeft == -1:
            if element.locationID not in get_all_empire_planets():
                # dequeue_list.append(queue_index) #TODO add assessment of recapture -- invasion target etc.
                debug(
                    "element %s will never be completed as stands and location %d no longer owned; could consider deleting from queue",
                    element.name,
                    element.locationID,
                )  # TODO:
            else:
                debug(
                    "element %s is projected to never be completed as currently stands, but will remain on queue ",
                    element.name,
                )
        elif element.buildType == EmpireProductionTypes.BT_SHIP:
            this_role = aistate.get_ship_role(element.designID)
            if this_role == ShipRoleType.CIVILIAN_COLONISATION:
                this_spec = universe.getPlanet(element.locationID).speciesName
                queued_colony_ships[this_spec] = (
                    queued_colony_ships.get(this_spec, 0) + element.remaining * element.blocksize
                )
            elif this_role == ShipRoleType.CIVILIAN_OUTPOST:
                queued_outpost_ships += element.remaining * element.blocksize
            elif this_role == ShipRoleType.BASE_OUTPOST:
                queued_outpost_ships += element.remaining * element.blocksize
            elif this_role == ShipRoleType.MILITARY_INVASION:
                queued_troop_ships += element.remaining * element.blocksize
            elif (this_role == ShipRoleType.CIVILIAN_EXPLORATION) and (queue_index <= 1):
                if len(AIstate.opponentPlanetIDs) > 0:
                    can_prioritize_troops = True
    if queued_colony_ships:
        debug("\nFound colony ships in build queue: %s", queued_colony_ships)
    if queued_outpost_ships:
        debug("\nFound outpost ships and bases in build queue: %s", queued_outpost_ships)

    for queue_index in dequeue_list[::-1]:
        fo.issueDequeueProductionOrder(queue_index)

    all_military_fleet_ids = FleetUtilsAI.get_empire_fleet_ids_by_role(MissionType.MILITARY)
    total_military_ships = sum([aistate.fleetStatus.get(fid, {}).get("nships", 0) for fid in all_military_fleet_ids])
    all_troop_fleet_ids = FleetUtilsAI.get_empire_fleet_ids_by_role(MissionType.INVASION)
    total_troop_ships = sum([aistate.fleetStatus.get(fid, {}).get("nships", 0) for fid in all_troop_fleet_ids])
    avail_troop_fleet_ids = list(FleetUtilsAI.extract_fleet_ids_without_mission_types(all_troop_fleet_ids))
    total_available_troops = sum([aistate.fleetStatus.get(fid, {}).get("nships", 0) for fid in avail_troop_fleet_ids])
    debug(
        "Trooper Status turn %d: %d total, with %d unassigned. %d queued, compared to %d total Military Attack Ships",
        current_turn,
        total_troop_ships,
        total_available_troops,
        queued_troop_ships,
        total_military_ships,
    )
    if (
        capital_id is not None
        and (current_turn >= 40 or can_prioritize_troops)
        and aistate.systemStatus.get(capital_system_id, {}).get("fleetThreat", 0) == 0
        and aistate.systemStatus.get(capital_system_id, {}).get("neighborThreat", 0) == 0
    ):
        best_design_id, best_design, build_choices = get_best_ship_info(PriorityType.PRODUCTION_INVASION)
        if build_choices is not None and len(build_choices) > 0:
            loc = random.choice(build_choices)
            prod_time = best_design.productionTime(empire.empireID, loc)
            prod_cost = best_design.productionCost(empire.empireID, loc)
            troopers_needed = max(
                0,
                int(
                    min(
                        0.99 + (current_turn / 20.0 - total_available_troops) / max(2, prod_time - 1),
                        total_military_ships // 3 - total_troop_ships,
                    )
                ),
            )
            ship_number = troopers_needed
            per_turn_cost = float(prod_cost) / prod_time
            if (
                troopers_needed > 0
                and empire.productionPoints > 3 * per_turn_cost * queued_troop_ships
                and aistate.character.may_produce_troops()
            ):
                retval = fo.issueEnqueueShipProductionOrder(best_design_id, loc)
                if retval != 0:
                    debug(
                        "forcing %d new ship(s) to production queue: %s; per turn production cost %.1f\n",
                        ship_number,
                        best_design.name,
                        ship_number * per_turn_cost,
                    )
                    if ship_number > 1:
                        fo.issueChangeProductionQuantityOrder(production_queue.size - 1, 1, ship_number)
                    avail_pp -= ship_number * per_turn_cost
                    fo.issueRequeueProductionOrder(production_queue.size - 1, 0)  # move to front
                    fo.updateProductionQueue()
        debug("")

    debug("")
    # get the highest production priorities
    production_priorities = {}
    for priority_type in get_priority_production_types():
        production_priorities[priority_type] = int(max(0, (aistate.get_priority(priority_type)) ** 0.5))

    sorted_priorities = sorted(production_priorities.items(), key=itemgetter(1), reverse=True)

    top_score = -1

    num_colony_fleets = len(
        FleetUtilsAI.get_empire_fleet_ids_by_role(MissionType.COLONISATION)
    )  # counting existing colony fleets each as one ship
    total_colony_fleets = sum(queued_colony_ships.values()) + num_colony_fleets
    num_outpost_fleets = len(
        FleetUtilsAI.get_empire_fleet_ids_by_role(MissionType.OUTPOST)
    )  # counting existing outpost fleets each as one ship
    total_outpost_fleets = queued_outpost_ships + num_outpost_fleets

    max_colony_fleets = PriorityAI.allottedColonyTargets
    max_outpost_fleets = max_colony_fleets

    _, _, colony_build_choices = get_best_ship_info(PriorityType.PRODUCTION_COLONISATION)
    military_emergency = PriorityAI.unmetThreat > (2.0 * MilitaryAI.get_tot_mil_rating())

    debug("Production Queue Priorities:")
    filtered_priorities = {}
    for priority_id, score in sorted_priorities:
        if military_emergency:
            if priority_id == PriorityType.PRODUCTION_EXPLORATION:
                score /= 10.0
            elif priority_id != PriorityType.PRODUCTION_MILITARY:
                score /= 2.0
        if top_score < score:
            top_score = score  # don't really need top_score nor sorting with current handling
        debug(" Score: %4d -- %s ", score, priority_id)
        if priority_id != PriorityType.PRODUCTION_BUILDINGS:
            if (
                (priority_id == PriorityType.PRODUCTION_COLONISATION)
                and (total_colony_fleets < max_colony_fleets)
                and (colony_build_choices is not None)
                and len(colony_build_choices) > 0
            ):
                filtered_priorities[priority_id] = score
            elif (priority_id == PriorityType.PRODUCTION_OUTPOST) and (total_outpost_fleets < max_outpost_fleets):
                filtered_priorities[priority_id] = score
            elif priority_id not in [PriorityType.PRODUCTION_OUTPOST, PriorityType.PRODUCTION_COLONISATION]:
                filtered_priorities[priority_id] = score
    if filtered_priorities == {}:
        debug("No non-building-production priorities with nonzero score, setting to default: Military")
        filtered_priorities[PriorityType.PRODUCTION_MILITARY] = 1
    if top_score <= 100:
        scaling_power = 1.0
    else:
        scaling_power = math.log(100) / math.log(top_score)
    for pty in filtered_priorities:
        filtered_priorities[pty] **= scaling_power

    available_pp = {
        tuple(el.key()): el.data() for el in empire.planetsWithAvailablePP
    }  # keys are sets of ints; data is doubles
    allocated_pp = {
        tuple(el.key()): el.data() for el in empire.planetsWithAllocatedPP
    }  # keys are sets of ints; data is doubles
    planets_with_wasted_pp = {tuple(pidset) for pidset in empire.planetsWithWastedPP}
    debug("avail_pp ( <systems> : pp ):")
    for planet_set in available_pp:
        debug(
            "\t%s\t%.2f",
            PlanetUtilsAI.sys_name_ids(set(PlanetUtilsAI.get_systems(planet_set))),
            available_pp[planet_set],
        )
    debug("\nallocated_pp ( <systems> : pp ):")
    for planet_set in allocated_pp:
        debug(
            "\t%s\t%.2f",
            PlanetUtilsAI.sys_name_ids(set(PlanetUtilsAI.get_systems(planet_set))),
            allocated_pp[planet_set],
        )

    debug("\n\nBuilding Ships in system groups with remaining PP:")
    for planet_set in planets_with_wasted_pp:
        total_pp = available_pp.get(planet_set, 0)
        avail_pp = total_pp - allocated_pp.get(planet_set, 0)
        if avail_pp <= 0.01:
            continue
        debug(
            "%.2f PP remaining in system group: %s",
            avail_pp,
            PlanetUtilsAI.sys_name_ids(set(PlanetUtilsAI.get_systems(planet_set))),
        )
        debug("\t owned planets in this group are:")
        debug("\t %s", PlanetUtilsAI.planet_string(planet_set))
        best_design_id, best_design, build_choices = get_best_ship_info(
            PriorityType.PRODUCTION_COLONISATION, planet_set
        )
        species_map = {}
        for loc in build_choices or []:
            this_spec = universe.getPlanet(loc).speciesName
            species_map.setdefault(this_spec, []).append(loc)
        colony_build_choices = []
        for pid, (score, this_spec) in aistate.colonisablePlanetIDs.items():
            # add planets multiple times to emulate choice with weight
            weight = int(math.ceil(score))
            planets_for_colonization = [pid_ for pid_ in species_map.get(this_spec, []) if pid_ in planet_set]
            weighted_planets = weight * planets_for_colonization
            colony_build_choices.extend(weighted_planets)

        local_priorities = {}
        local_priorities.update(filtered_priorities)
        best_ships = {}
        mil_build_choices = get_best_ship_ratings(planet_set)
        for priority in list(local_priorities):
            if priority == PriorityType.PRODUCTION_MILITARY:
                if not mil_build_choices:
                    del local_priorities[priority]
                    continue
                _, pid, best_design_id, best_design = mil_build_choices[0]
                build_choices = [pid]
                # score = ColonisationAI.pilotRatings.get(pid, 0)
                # if bestScore < ColonisationAI.curMidPilotRating:
            else:
                best_design_id, best_design, build_choices = get_best_ship_info(priority, planet_set)
            if best_design is None:
                del local_priorities[priority]  # must be missing a shipyard -- TODO build a shipyard if necessary
                continue
            best_ships[priority] = [best_design_id, best_design, build_choices]
            debug("best_ships[%s] = %s \t locs are %s from %s", priority, best_design.name, build_choices, planet_set)

        if len(local_priorities) == 0:
            debug(
                "Alert!! need shipyards in systemSet %s",
                PlanetUtilsAI.sys_name_ids(set(PlanetUtilsAI.get_systems(planet_set))),
            )
        priority_choices = []
        for priority in local_priorities:
            priority_choices.extend(int(local_priorities[priority]) * [priority])

        loop_count = 0
        while (
            (avail_pp > 0) and (loop_count < max(100, current_turn)) and (priority_choices != [])
        ):  # make sure don't get stuck in some nonbreaking loop like if all shipyards captured
            loop_count += 1
            debug("Beginning build enqueue loop %d; %.1f PP available", loop_count, avail_pp)
            this_priority = random.choice(priority_choices)
            debug("selected priority: %s", this_priority)
            making_colony_ship = False
            making_outpost_ship = False
            if this_priority == PriorityType.PRODUCTION_COLONISATION:
                if total_colony_fleets >= max_colony_fleets:
                    debug("Already sufficient colony ships in queue, trying next priority choice\n")
                    for i in range(len(priority_choices) - 1, -1, -1):
                        if priority_choices[i] == PriorityType.PRODUCTION_COLONISATION:
                            del priority_choices[i]
                    continue
                elif colony_build_choices is None or len(colony_build_choices) == 0:
                    for i in range(len(priority_choices) - 1, -1, -1):
                        if priority_choices[i] == PriorityType.PRODUCTION_COLONISATION:
                            del priority_choices[i]
                    continue
                else:
                    making_colony_ship = True
            if this_priority == PriorityType.PRODUCTION_OUTPOST:
                if total_outpost_fleets >= max_outpost_fleets:
                    debug("Already sufficient outpost ships in queue, trying next priority choice\n")
                    for i in range(len(priority_choices) - 1, -1, -1):
                        if priority_choices[i] == PriorityType.PRODUCTION_OUTPOST:
                            del priority_choices[i]
                    continue
                else:
                    making_outpost_ship = True
            best_design_id, best_design, build_choices = best_ships[this_priority]
            if making_colony_ship:
                loc = random.choice(colony_build_choices)
                best_design_id, best_design, build_choices = get_best_ship_info(
                    PriorityType.PRODUCTION_COLONISATION, loc
                )
            elif this_priority == PriorityType.PRODUCTION_MILITARY:
                selector = random.random()
                choice = mil_build_choices[0]  # mil_build_choices can't be empty due to earlier check
                for choice in mil_build_choices:
                    if choice[0] >= selector:
                        break
                loc, best_design_id, best_design = choice[1:4]
                if best_design is None:
                    warning(
                        "problem with mil_build_choices;"
                        f" with selector ({selector}) chose loc ({loc}), "
                        f"best_design_id ({best_design_id}), best_design (None) "
                        f"from mil_build_choices: {mil_build_choices}"
                    )
                    continue
            else:
                loc = random.choice(build_choices)

            ship_number = 1
            per_turn_cost = float(best_design.productionCost(empire.empireID, loc)) / best_design.productionTime(
                empire.empireID, loc
            )
            if this_priority == PriorityType.PRODUCTION_MILITARY:
                this_rating = get_rating_for_planet(pid)
                rating_ratio = float(this_rating) / best_pilot_rating()
                if rating_ratio < 0.1:
                    loc_planet = universe.getPlanet(loc)
                    if loc_planet:
                        pname = loc_planet.name
                        this_rating = rate_planetary_piloting(loc)
                        rating_ratio = float(this_rating) / best_pilot_rating()
                        qualifier = "suboptimal " if rating_ratio < 1.0 else ""
                        debug(
                            "Building mil ship at loc %d (%s) with %spilot Rating: %.1f; ratio to empire best is %.1f",
                            loc,
                            pname,
                            qualifier,
                            this_rating,
                            rating_ratio,
                        )
                while total_pp > 40 * per_turn_cost:
                    ship_number *= 2
                    per_turn_cost *= 2
            retval = fo.issueEnqueueShipProductionOrder(best_design_id, loc)
            if retval != 0:
                prioritized = False
                debug(
                    "adding %d new ship(s) at location %s to production queue: %s; per turn production cost %.1f\n",
                    ship_number,
                    PlanetUtilsAI.planet_string(loc),
                    best_design.name,
                    per_turn_cost,
                )
                if ship_number > 1:
                    fo.issueChangeProductionQuantityOrder(production_queue.size - 1, 1, ship_number)
                avail_pp -= per_turn_cost
                if making_colony_ship:
                    total_colony_fleets += ship_number
                    if total_pp > 4 * per_turn_cost:
                        fo.issueRequeueProductionOrder(production_queue.size - 1, 0)  # move to front
                    continue
                if making_outpost_ship:
                    total_outpost_fleets += ship_number
                    if total_pp > 4 * per_turn_cost:
                        fo.issueRequeueProductionOrder(production_queue.size - 1, 0)  # move to front
                    continue
                if total_pp > 10 * per_turn_cost:
                    leading_block_pp = 0
                    for elem in [production_queue[elemi] for elemi in range(0, min(4, production_queue.size))]:
                        cost, time = empire.productionCostAndTime(elem)
                        leading_block_pp += elem.blocksize * cost / time
                    if leading_block_pp > 0.5 * total_pp or (
                        military_emergency and this_priority == PriorityType.PRODUCTION_MILITARY
                    ):
                        prioritized = True
                        fo.issueRequeueProductionOrder(production_queue.size - 1, 0)  # move to front
                if this_priority == PriorityType.PRODUCTION_INVASION:
                    queued_troop_ships += ship_number
                    if not prioritized:
                        fo.issueRequeueProductionOrder(production_queue.size - 1, 0)  # move to front
        # The AI will normally only consider queuing additional ships (above) the current queue is not using all the
        # available PP; this can delay the AI from pursuing easy invasion targets.
        # Queue an extra troopship if the following conditions are met:
        #   i) currently dealing with our Capital ResourceGroup
        #   ii) the invasion priority for this group is nonzero and the max priority, and
        #   iii) there are minimal troopships already enqueued
        invasion_priority = local_priorities.get(PriorityType.PRODUCTION_INVASION, 0)
        if (
            capital_id in planet_set
            and invasion_priority
            and invasion_priority == max(local_priorities.values())
            and queued_troop_ships <= 2
        ):  # todo get max from character module or otherwise calculate
            best_design_id, best_design, build_choices = best_ships[PriorityType.PRODUCTION_INVASION]
            loc = random.choice(build_choices)
            retval = fo.issueEnqueueShipProductionOrder(best_design_id, loc)
            if retval != 0:
                per_turn_cost = float(best_design.productionCost(empire.empireID, loc)) / best_design.productionTime(
                    empire.empireID, loc
                )
                avail_pp -= per_turn_cost
                debug(
                    "adding extra trooper at location %s to production queue: %s; per turn production cost %.1f\n",
                    PlanetUtilsAI.planet_string(loc),
                    best_design.name,
                    per_turn_cost,
                )

        debug("")
    update_stockpile_use()
    fo.updateProductionQueue()
    print_production_queue(after_turn=True)


def _get_queued_buildings(pid: PlanetId) -> list[BuildingName]:
    debug("Buildings already in Production Queue:")
    capital_queued_buildings = _get_queued_buildings_for_planet(pid)
    for bldg in capital_queued_buildings:
        debug(f"    {bldg.name} turns: {bldg.turnsLeft} PP: {bldg.allocation}")
    if not capital_queued_buildings:
        debug("No capital queued buildings")
    queued_building_names = [bldg.name for bldg in capital_queued_buildings]
    return queued_building_names


def _is_queued_building_on_planet(e: "fo.productionQueueElement", pid: PlanetId) -> bool:
    return e.buildType == EmpireProductionTypes.BT_BUILDING and e.locationID == pid


def _get_queued_buildings_for_planet(pid: PlanetId) -> Sequence["fo.productionQueueElement"]:
    queue = fo.getEmpire().productionQueue
    return [e for e in queue if _is_queued_building_on_planet(e, pid)]


def update_stockpile_use():
    """Decide which elements in the production_queue will be enabled for drawing from the imperial stockpile.  This
    initial version simply ensures that every resource group with at least one item on the queue has its highest
    priority item be stockpile-enabled.

    :return: None
    """
    # TODO: Do a priority and risk evaluation to decide on enabling stockpile draws
    empire = fo.getEmpire()
    production_queue = empire.productionQueue
    resource_groups = {tuple(el.key()) for el in empire.planetsWithAvailablePP}
    planets_in_stockpile_enabled_group = set()
    for queue_index, element in enumerate(production_queue):
        if element.locationID in planets_in_stockpile_enabled_group:
            # TODO: evaluate possibly disabling stockpile for current element if was previously enabled, perhaps
            # only allowing multiple stockpile enabled items in empire-capital-resource-group, or considering some
            # priority analysis
            continue
        group = next((_g for _g in resource_groups if element.locationID in _g), None)
        if group is None:
            continue  # we don't appear to own the location any more
        if fo.issueAllowStockpileProductionOrder(queue_index, True):
            planets_in_stockpile_enabled_group.update(group)


def empire_has_colony_bld_species(building_name: str) -> bool:
    """
    Checks if this building is a colony building for which this empire has the required source species available.
    """
    if not building_name.startswith("BLD_COL_"):
        return False
    species_name = "SP_" + building_name.split("BLD_COL_")[1]
    return can_build_colony_for_species(species_name)


def already_has_completed_colony_building(planet_id) -> bool:
    """
    Checks if a planet has an already-completed (but not yet 'hatched') colony building.
    """
    universe = fo.getUniverse()
    planet = universe.getPlanet(planet_id)
    return any(universe.getBuilding(bldg).name.startswith("BLD_COL_") for bldg in planet.buildingIDs)


def _build_ship_facilities(building_type: Shipyard, top_pids: set[PlanetId] = frozenset()) -> None:
    # TODO: add total_pp checks below, so don't overload queue
    if not building_type.available():
        return
    universe = fo.getUniverse()
    total_pp = fo.getEmpire().productionPoints
    prerequisite_type = building_type.prerequisite()
    queued_bld_pids = building_type.queued_at()
    if building_type in Shipyard.get_system_ship_facilities():
        current_coverage = building_type.built_or_queued_at_sys()
        open_systems = {
            universe.getPlanet(pid).systemID for pid in get_best_pilot_facilities(Shipyard.BASE.value)
        }.difference(current_coverage)
        try_systems = open_systems & prerequisite_type.built_or_queued_at_sys() if prerequisite_type else open_systems
        try_pids = {pid for sys_id in try_systems for pid in get_owned_planets_in_system(sys_id)}
    else:
        current_pids = get_best_pilot_facilities(building_type.value)
        try_pids = get_best_pilot_facilities(prerequisite_type.value).difference(queued_bld_pids, current_pids)
    debug(
        "Considering constructing a %s, have %d already built and %d queued",
        building_type,
        len(building_type.built_at()),
        len(building_type.queued_at()),
    )
    if not try_pids:
        # Recursion cannot handle the base yard, but currently the AI builds it everywhere anyway.
        # Production of shipyards should be re-written, the AI builds too many of them.
        if prerequisite_type != Shipyard.BASE:
            debug(f"Cannot build {building_type} at top-pilot planets, try building {prerequisite_type} first.")
            _build_ship_facilities(prerequisite_type, top_pids)
        return
    # ship facilities all have location independent costs
    turn_cost = building_type.turn_cost(list(try_pids)[0])
    max_under_construction = max(1, int(total_pp) // (int(5 * turn_cost)))
    max_total = max(1, int(total_pp) // int(2 * turn_cost))
    debug("Allowances: max total: %d, max under construction: %d", max_total, max_under_construction)
    if len(building_type.built_at()) >= max_total:
        return
    try_top_pids = [pid for pid in try_pids & top_pids if building_type.can_be_produced(pid)]
    try_other_pids = [pid for pid in try_pids - top_pids if building_type.can_be_produced(pid)]
    valid_pids = try_top_pids + try_other_pids
    debug("Have %d potential locations: %s", len(valid_pids), [universe.getPlanet(x) for x in valid_pids])
    # TODO: rank by defense ability, etc.
    num_queued = len(queued_bld_pids)
    already_covered = []  # just those covered on this turn
    while valid_pids:
        if num_queued >= max_under_construction:
            break
        pid = valid_pids.pop()
        if pid in already_covered:
            continue
        res = building_type.enqueue(pid)
        debug("Enqueueing %s at planet %s , with result %d", building_type, universe.getPlanet(pid), res)
        if res:
            num_queued += 1
            already_covered.extend(get_owned_planets_in_system(universe.getPlanet(pid).systemID))


def find_automatic_historic_analyzer_candidates() -> list[int]:
    """
    Find possible locations for the BLD_AUTO_HISTORY_ANALYSER building and return a subset of chosen building locations.

    :return: Random possible locations up to max queueable amount. Empty if no location found or can't queue another one
    """
    empire = fo.getEmpire()
    universe = fo.getUniverse()
    total_pp = empire.productionPoints
    history_analyser = "BLD_AUTO_HISTORY_ANALYSER"
    culture_archives = "BLD_CULTURE_ARCHIVES"

    ARB_LARGE_NUMBER = 1e4
    conditions = {
        # aggression: (min_pp, min_turn, min_pp_to_queue_another_one)
        fo.aggression.beginner: (100, 100, ARB_LARGE_NUMBER),
        fo.aggression.turtle: (75, 75, ARB_LARGE_NUMBER),
        fo.aggression.cautious: (40, 40, ARB_LARGE_NUMBER),
        fo.aggression.typical: (20, 20, 50),
        fo.aggression.aggressive: (10, 10, 40),
        fo.aggression.maniacal: (8, 5, 30),
    }

    min_pp, turn_trigger, min_pp_per_additional = conditions.get(
        get_aistate().character.get_trait(Aggression).key, (ARB_LARGE_NUMBER, ARB_LARGE_NUMBER, ARB_LARGE_NUMBER)
    )
    max_enqueued = 1 if total_pp > min_pp or fo.currentTurn() > turn_trigger else 0
    max_enqueued += int(total_pp / min_pp_per_additional)

    if max_enqueued <= 0:
        return []

    # find possible locations
    possible_locations = set()
    for pid in get_all_empire_planets():
        planet = universe.getPlanet(pid)
        if not planet or planet.currentMeterValue(fo.meterType.targetPopulation) < 1:
            continue
        buildings_here = [bld.buildingTypeName for bld in map(universe.getBuilding, planet.buildingIDs)]
        if planet and culture_archives in buildings_here and history_analyser not in buildings_here:
            possible_locations.add(pid)

    # check existing queued buildings and remove from possible locations
    queued_locs = {
        e.locationID
        for e in empire.productionQueue
        if e.buildType == EmpireProductionTypes.BT_BUILDING and e.name == history_analyser
    }

    possible_locations -= queued_locs
    chosen_locations = []
    for i in range(min(max_enqueued, len(possible_locations))):
        chosen_locations.append(possible_locations.pop())
    return chosen_locations


def _location_rating(planet: fo.planet) -> float:
    """Get a rating how good this planet would be for a building"""
    # Simple value so far, should be enhanced, e.g. a scanner should go to the planet with species that
    # has the best basic visions
    return planet.currentMeterValue(fo.meterType.maxTroops)


def _try_enqueue(
    building_type: BuildingTypeBase,
    candidates: Union[PlanetId, Iterable[PlanetId]],
    *,
    at_front: bool = False,
    ignore_dislike: bool = False,
) -> float:
    """
    Enqueue building at one of the planets in candidates.
    If at_front, building is added at the front of the queue.
    Returns PP per turn spent on the new building or 0.0 if nothing was enqueued.
    """
    universe = fo.getUniverse()
    empire = fo.getEmpire()
    opinion = building_type.get_opinions()
    locations = []
    preferred_locations = []
    if isinstance(candidates, int):  # isinstance(candidates, PlanetId) does not work, at least not in python-3.9
        candidates = [candidates]
    for pid in candidates:
        planet = universe.getPlanet(pid)
        if not planet:
            error(f"Got pid {pid} in candidate, which does not seem to be a planetID")
            continue
        if not ignore_dislike and pid in opinion.dislikes:
            continue
        if not building_type.can_be_produced(pid) or not building_type.can_be_enqueued(pid):
            continue
        if pid in opinion.likes:
            preferred_locations.append((_location_rating(planet), pid))
        else:
            locations.append((_location_rating(planet), pid))
    for _, pid in sorted(preferred_locations, reverse=True) + sorted(locations, reverse=True):
        planet = universe.getPlanet(pid)
        res = building_type.enqueue(pid)
        debug("Enqueueing %s at planet %d (%s) , with result %d", building_type, pid, planet.name, res)
        if res:
            if at_front:
                res = fo.issueRequeueProductionOrder(empire.productionQueue.size - 1, 0)  # move to front
                debug("Requeueing %s to front of build queue, with result %d", building_type, res)
            return building_type.turn_cost(pid)
    return 0.0


def _may_enqueue_for_stability(building_type: BuildingTypeBase, new_turn_cost: float) -> float:
    """
    Build building if it seems worth doing so to increase stability.
    Only builds of locations.planets_enqueued is empty and new_turn_cost is 0.0,
    i.e. there are currently no build queue entries for the given building.
    returns new_turn_cost or turn_cost of the building enqueued by this function.
    """
    if building_type.queued_at() or new_turn_cost:
        return new_turn_cost
    # this can be improved a lot, taking into account value of planets, actual stability and
    # what effects the change would have. For the moment, keep it simple.
    # Note that the strongest effect is always on the building's planet itself.
    opinion = building_type.get_opinions()
    universe = fo.getUniverse()
    if len(opinion.likes) >= len(opinion.dislikes) * PlanetUtilsAI.dislike_factor():
        like_candidates = opinion.likes - building_type.built_or_queued_at()
        # plans may change, so consider only actual colonies that like it
        candidates = [pid for pid in like_candidates if universe.getPlanet(pid).speciesName]
        return _try_enqueue(building_type, candidates)
    return 0.0


def _build_scanning_facility() -> float:
    """Consider building Scanning Facilities, return added turn costs."""
    building_type = BuildingType.SCANNING_FACILITY
    empire = fo.getEmpire()
    if not building_type.available():
        return 0.0

    universe = fo.getUniverse()
    turn_cost = 0.0
    opinion = building_type.get_opinions()
    # TBD use actual cost?
    max_scanner_builds = max(1, int(empire.productionPoints / 30)) - len(building_type.queued_at())
    scanner_systems = building_type.built_or_queued_at_sys()
    debug(
        "Considering building %s, found current and queued systems %s, planets that like it %s, #dislikes: %d",
        building_type,
        PlanetUtilsAI.sys_name_ids(scanner_systems),
        PlanetUtilsAI.sys_name_ids(opinion.likes),
        len(opinion.dislikes),
    )
    for sys_id in get_owned_planets():
        if max_scanner_builds <= 0:
            break
        if sys_id in scanner_systems:
            continue
        need_scanner = False
        for neighbor in get_neighbors(sys_id):
            if universe.getVisibility(neighbor, empire.empireID) < fo.visibility.partial:
                need_scanner = True
                break
        if not need_scanner:
            continue
        # TBD: chose based on detection range
        cost = _try_enqueue(building_type, get_owned_planets_in_system(sys_id), at_front=True)
        if cost:
            max_scanner_builds -= 1
        turn_cost += cost
    return _may_enqueue_for_stability(building_type, turn_cost)


def _build_gas_giant_generator() -> float:  # noqa: C901
    """Consider building Gas Giant Generators, return added turn costs."""
    building_type = BuildingType.GAS_GIANT_GEN
    if not building_type.available():
        return 0.0

    ggg_min_stability = get_named_real("BLD_GAS_GIANT_GEN_MIN_STABILITY")
    universe = fo.getUniverse()
    colonized_planets = get_colonized_planets()
    opinion = building_type.get_opinions()
    systems = []
    for sys in colonized_planets.keys():
        planets = [(pid, universe.getPlanet(pid)) for pid in get_owned_planets_in_system(sys)]
        if sys in building_type.built_or_queued_at_sys() or fo.planetSize.gasGiant not in [x[1].size for x in planets]:
            continue
        rating = 0
        gas_giant = None
        best_gg = -2
        debug(f"Gas Giant Generator rating for {universe.getSystem(sys).name} ...")
        for pid, planet in planets:
            likes = opinion.value(pid, 1, 0, -1 * PlanetUtilsAI.dislike_factor())
            debug(f"  {planet.name} likes {likes}")
            # TBD -4 if build here...
            stability = planet.currentMeterValue(fo.meterType.targetHappiness) + likes
            rating += 3 * likes
            debug(f"  rating now {rating} from likes {likes} ")
            if planet.size == fo.planetSize.gasGiant:
                val = likes
                if val > best_gg:
                    best_gg = val
                    gas_giant = pid
            if planet.focus == FocusType.FOCUS_INDUSTRY and stability >= ggg_min_stability:
                rating += 20 + min(5, stability - ggg_min_stability)
                debug(f"  rating now {rating} from industry planet stability {stability} ")
            elif FocusType.FOCUS_INDUSTRY in planet.availableFoci and stability >= ggg_min_stability:
                rating += 5 + min(5, stability - ggg_min_stability)
                debug(f"  rating now {rating} from pot. industry planet stability {stability} ")
        if gas_giant:
            # if the inhabitants do not like it, this will require two other planets that profit from it
            rating += 15 * best_gg
            debug(f"  from best_gg {best_gg}, final rating: {rating}")
            systems.append((rating, gas_giant))
    # sorting so that highest ratings come last, which means they end up at the front of the queue
    systems.sort()
    turn_cost = 0.0
    for rating, gas_giant in systems:
        # 20 = one industry planet with exactly ggg_min_stability
        if rating >= 20:
            turn_cost += _try_enqueue(building_type, gas_giant, at_front=True, ignore_dislike=True)
    return _may_enqueue_for_stability(building_type, turn_cost)


def _build_translator():
    """Consider building Near Universal Translators, return added turn costs."""
    building_type = BuildingType.TRANSLATOR
    if building_type.available() and translators_wanted() and candidate_for_translator:
        # starting one per turn should be enough
        have_one = bool(building_type.built_or_queued_at())
        return _try_enqueue(building_type, candidate_for_translator, at_front=not have_one)
    return 0.0
    # may_enqueue_for_stability? Building is rather expensive...


def _build_regional_administration() -> float:
    """Consider building Imperial Regional Administrations, return added turn costs."""
    building_type = BuildingType.REGIONAL_ADMIN
    current_admin_systems = building_type.built_or_queued_at_sys() | BuildingType.PALACE.built_or_queued_at_sys()
    # No administrations at all means no palace. If we cannot even find a place for the palace,
    # there is no point in building regional administrations.
    if not building_type.available() or not current_admin_systems:
        return 0.0
    universe = fo.getUniverse()
    jumps_to_admin = [
        (min(universe.jumpDistance(sys_id, admin) for admin in current_admin_systems), sys_id)
        for sys_id in get_owned_planets()
    ]
    jumps_to_admin.sort(reverse=True)
    # Currently, 6 is required, this is supposed to change. Note that the game allows to enqueue several ones
    # close to each other, but only one will get finished.
    if jumps_to_admin[0][0] < 6:
        return 0.0
    debug(f"current_admin_systems: {current_admin_systems}, jumps_to_admin = {jumps_to_admin}")
    # with minimum distance 6, systems 3 jumps away from current admins cannot get closer
    systems_that_may_profit = [value for value in jumps_to_admin if value[0] > 3]
    best_sys_id = None
    # without a threshold the AI would build at the first planet 6 steps away from others,
    # although it may not give much and is possible quite exposed.
    best_rating = 10.0
    for distance, candidate in jumps_to_admin:
        if distance < 6:
            break
        rating = _rate_system_for_admin(candidate, systems_that_may_profit)
        if rating > best_rating:
            best_sys_id = candidate
            best_rating = rating
    # To add more than one, we'd have to recalculate everything, but one per turn is good enough.
    # In practice more than one would hardly ever be possible anyway.
    debug(f"best_sys_id={best_sys_id}, best_rating={best_rating}, planets: {get_owned_planets_in_system(best_sys_id)}")
    if not best_sys_id:
        return 0.0
    return _try_enqueue(building_type, get_owned_planets_in_system(best_sys_id))


def _rate_system_for_admin(sys_id: SystemId, systems_that_may_profit: list[tuple[int, SystemId]]) -> float:
    opinion = BuildingType.REGIONAL_ADMIN.get_opinions()
    planets = set(get_owned_planets_in_system(sys_id))
    dislikes = planets & opinion.dislikes
    likes = planets & opinion.likes
    if dislikes == planets:
        return 0.0
    # First like gets a big bonus, but we can build it only on one.
    # Number of planets to prefer better defended systems.
    rating = 1.5 * (len(likes) - len(dislikes) * PlanetUtilsAI.dislike_factor()) + 3 * (likes != set()) + len(planets)

    universe = fo.getUniverse()
    for current_distance, other_sys_id in systems_that_may_profit:
        difference = current_distance - universe.jumpDistance(sys_id, other_sys_id)
        if difference > 0:
            for pid in get_colonized_planets_in_system(other_sys_id):
                planet = universe.getPlanet(pid)
                if Tags.INDEPENDENT not in fo.getSpecies(planet.speciesName).tags:
                    stability = planet.currentMeterValue(fo.meterType.targetHappiness)
                    rating += difference
                    # reaching 10 gives a lot of bonuses
                    if stability < 0 or stability < 10 <= stability + difference:
                        rating += 2
    debug(f"admin rating {universe.getSystem(sys_id)}={rating}")
    return rating


def _build_military_command() -> float:
    """
    Consider building a Military Command, return added turn costs.
    Since its major purpose is to provide policy slots, and we may need the production for more important
    things, do not build it too early. Won't build it at all, if all our planets dislike it.
    """
    building_type = BuildingType.MILITARY_COMMAND
    palace_planet = BuildingType.PALACE.built_at()
    # an empire can only build one, and if we do not have a palace, this is definitely more important
    if building_type.built_or_queued_at() or not palace_planet:
        return 0.0
    # cost is independent of the location, but we need a valid location
    if fo.getEmpire().productionPoints > building_type.turn_cost(list(palace_planet)[0]) * 1.5:
        # default selection should prefer the capital, unless its species dislikes it.
        return _try_enqueue(building_type, get_inhabited_planets())
    return 0.0


TopPilotSystems = NewType("TopPilotSystems", dict[SystemId, list[tuple[PlanetId, float]]])


class ShipYardInfo(NamedTuple):
    queued_shipyard_pids: list[PlanetId]
    colony_systems: dict[PlanetId, SystemId]
    top_pilot_systems: TopPilotSystems


def _build_basic_shipyards() -> ShipYardInfo:  # noqa: C901
    """
    Consider building basic ship yards and also determine some value needed for other shipyard buildings.
    """
    building_type = Shipyard.BASE
    universe = fo.getUniverse()
    queued_shipyard_pids = building_type.queued_at()
    system_colonies = {}
    colony_systems = {}
    empire_species = get_empire_planets_by_species()
    for spec_name in get_colony_builders():
        if not get_colony_builder_locations(spec_name) and (
            spec_name in empire_species
        ):  # not enough current shipyards for this species #TODO: also allow orbital incubators and/or asteroid ships
            for pid in get_empire_planets_with_species(
                spec_name
            ):  # SP_EXOBOT may not actually have a colony yet but be in empireColonizers
                if pid in queued_shipyard_pids:
                    break  # won't try building more than one shipyard at once, per colonizer
            else:
                # no queued shipyards: get planets with target pop >=3 and
                # queue a shipyard on the one with the biggest current population
                planets = (universe.getPlanet(x) for x in get_empire_planets_with_species(spec_name))
                pops = sorted(
                    (planet_.initialMeterValue(fo.meterType.population), planet_.id)
                    for planet_ in planets
                    if (planet_ and planet_.initialMeterValue(fo.meterType.targetPopulation) >= 3.0)
                )
                pids = [pid for pop, pid in pops if building_type.can_be_produced(pid)]
                if pids:
                    build_loc = pids[-1]
                    res = _try_enqueue(building_type, build_loc)  # do not ignore dislikes here
                    if res > 0:
                        queued_shipyard_pids.append(build_loc)
                        break  # only start at most one new shipyard per species per turn
        for pid in get_empire_planets_with_species(spec_name):
            planet = universe.getPlanet(pid)
            if planet:
                system_colonies.setdefault(planet.systemID, {}).setdefault("pids", []).append(pid)
                colony_systems[pid] = planet.systemID

    for pid in get_empire_planets_with_species("SP_ACIREMA"):
        if (pid in queued_shipyard_pids) or not building_type.can_be_produced(pid):
            continue  # but not 'break' because we want to build shipyards at *every* Acirema planet
        # currently Acirema do not dislike ship yards, but if that changes, do not build shipyards anymore
        res = _try_enqueue(building_type, pid, at_front=True)
        if res > 0:
            queued_shipyard_pids.append(pid)

    top_pilot_systems = TopPilotSystems({})
    for pid, rating in get_pilot_ratings().items():
        if (rating <= medium_pilot_rating()) and (rating < GREAT_PILOT_RATING):
            continue
        top_pilot_systems.setdefault(universe.getPlanet(pid).systemID, []).append((pid, rating))
        if (pid in queued_shipyard_pids) or not building_type.can_be_produced(pid):
            continue  # but not 'break' because we want to build shipyards all top pilot planets
        # so far we ignore dislikes here, but this may have to change for Mu Ursh
        res = _try_enqueue(building_type, pid, at_front=True, ignore_dislike=True)
        if res:
            queued_shipyard_pids.append(pid)
    return ShipYardInfo(queued_shipyard_pids, colony_systems, top_pilot_systems)  # TBD return added costs?


def _build_energy_shipyards(  # noqa: C901
    queued_shipyard_pids: list[PlanetId],
    colony_systems: dict[PlanetId, SystemId],
    building_ratio: float,
    building_expense: float,
) -> tuple[list[tuple[float, PlanetId]], list[tuple[float, PlanetId]], float]:
    """
    Consider building Energy Compressor and Solar Containment Unit.
    Also determines pilot rating for planets in system with red stars and black holes.
    Returns blackhole_pilots, red_pilots and new value of building_expense.
    """
    universe = fo.getUniverse()
    empire = fo.getEmpire()
    pop_ctrs = list(get_inhabited_planets())
    red_population_centres = sorted(
        [
            (get_rating_for_planet(pid), pid)
            for pid in pop_ctrs
            if colony_systems.get(pid, INVALID_ID) in AIstate.empireStars.get(fo.starType.red, [])
        ],
        reverse=True,
    )
    red_pilots = [pid for _, pid in red_population_centres if _ == best_pilot_rating()]
    blue_population_centres = sorted(
        [
            (get_rating_for_planet(pid), pid)
            for pid in pop_ctrs
            if colony_systems.get(pid, INVALID_ID) in AIstate.empireStars.get(fo.starType.blue, [])
        ],
        reverse=True,
    )
    blue_pilots = [pid for _, pid in blue_population_centres if _ == best_pilot_rating()]
    blackhole_pilots = sorted(
        [
            (get_rating_for_planet(pid), pid)
            for pid in pop_ctrs
            if colony_systems.get(pid, INVALID_ID) in AIstate.empireStars.get(fo.starType.blackHole, [])
        ],
        reverse=True,
    )
    blackhole_pilots = [pid for _, pid in blackhole_pilots if _ == best_pilot_rating()]
    energy_shipyard_pids = {}
    building_type = Shipyard.ENRG_COMP
    if building_type.available():
        queued_building_pids = building_type.queued_at()
        for pid in blackhole_pilots + blue_pilots:
            if len(queued_building_pids) > 1:  # build a max of 2 at once
                break
            this_planet = universe.getPlanet(pid)
            if not (
                this_planet and can_build_ship_for_species(this_planet.speciesName)
            ):  # TODO: also check that not already one for this spec in this sys
                continue
            energy_shipyard_pids.setdefault(this_planet.systemID, []).append(pid)
            if pid not in queued_building_pids and building_type.can_be_produced(pid):
                building_expense += _try_enqueue(building_type, pid, at_front=True)

    building_type = Shipyard.ENRG_SOLAR
    if building_type.available() and not building_type.queued_at():
        # TODO: check that production is not frozen at a queued location
        for pid in blackhole_pilots:
            this_planet = universe.getPlanet(pid)
            if not (
                this_planet and can_build_ship_for_species(this_planet.speciesName)
            ):  # TODO: also check that not already one for this spec in this sys
                continue
            if building_type.can_be_produced(pid):
                building_expense += _try_enqueue(building_type, pid, at_front=True)

    total_pp = empire.productionPoints
    building_type = Shipyard.BASE
    if building_type.available() and (building_expense < building_ratio * total_pp) and (total_pp > 50):
        for sys_id in energy_shipyard_pids:  # Todo ensure only one or 2 per sys
            # only start one per turn (TBD why [:2]?)
            for pid in energy_shipyard_pids[sys_id][:2]:
                res = _try_enqueue(building_type, pid, at_front=True)
                if res > 0:
                    queued_shipyard_pids.append(pid)
                    break  # only start one per turn
    return blackhole_pilots, red_pilots, building_expense


def _build_asteroid_processor(  # noqa: C901
    top_pilot_systems: TopPilotSystems, queued_shipyard_pids: list[PlanetId]
) -> float:
    """Consider building asteroid processor, return added turn costs."""
    building_type = Shipyard.ASTEROID
    building_expense = 0.0
    if building_type.available():
        universe = fo.getUniverse()
        queued_building_pids = building_type.queued_at()
        if not queued_building_pids:
            asteroid_systems = {}
            asteroid_yards = {}
            builder_systems = {}
            for pid in get_all_empire_planets():
                planet = universe.getPlanet(pid)
                this_spec = planet.speciesName
                sys_id = planet.systemID
                if planet.size == fo.planetSize.asteroids and sys_id in get_colonized_planets():
                    asteroid_systems.setdefault(sys_id, []).append(pid)
                    if pid in building_type.built_or_queued_at():
                        asteroid_yards[sys_id] = pid  # shouldn't ever overwrite another, but ok if it did
                if can_build_ship_for_species(this_spec):
                    if pid not in get_ship_builder_locations(this_spec):
                        builder_systems.setdefault(sys_id, []).append((planet.speciesName, pid))
            # check if we need to build another asteroid processor:
            # check if local shipyard to go with the asteroid processor
            yard_systems = []
            need_yard = {}
            top_pilot_locations = []
            for sys_id in set(asteroid_systems.keys()).difference(asteroid_yards.keys()):
                if sys_id in top_pilot_systems:
                    for pid, rating in top_pilot_systems[sys_id]:
                        if pid not in queued_shipyard_pids:  # will catch it later if shipyard already present
                            top_pilot_locations.append((rating, pid, sys_id))
            top_pilot_locations.sort(reverse=True)
            for _, _, sys_id in top_pilot_locations:
                if sys_id not in yard_systems:
                    yard_systems.append(sys_id)  # prioritize asteroid yards for acirema and/or other top pilots
                    for pid, _ in top_pilot_systems[sys_id]:
                        if pid not in queued_shipyard_pids:  # will catch it later if shipyard already present
                            need_yard[sys_id] = pid
            if (not yard_systems) and len(asteroid_yards.values()) <= int(
                fo.currentTurn() // 50
            ):  # not yet building & not enough current locs, find a location to build one
                colonizer_loc_choices = []
                builder_loc_choices = []
                bld_systems = set(asteroid_systems.keys()).difference(asteroid_yards.keys())
                for sys_id in bld_systems.intersection(builder_systems.keys()):
                    for this_spec, pid in builder_systems[sys_id]:
                        if can_build_colony_for_species(this_spec):
                            if pid in (get_colony_builder_locations(this_spec) + queued_shipyard_pids):
                                colonizer_loc_choices.insert(0, sys_id)
                            else:
                                colonizer_loc_choices.append(sys_id)
                                need_yard[sys_id] = pid
                        else:
                            if pid in (get_ship_builder_locations(this_spec) + queued_shipyard_pids):
                                builder_loc_choices.insert(0, sys_id)
                            else:
                                builder_loc_choices.append(sys_id)
                                need_yard[sys_id] = pid
                yard_systems.extend(
                    (colonizer_loc_choices + builder_loc_choices)[:1]
                )  # add at most one of these non top pilot locs
            new_yard_count = len(queued_building_pids)
            for sys_id in yard_systems:  # build at most 2 new asteroid yards at a time
                if new_yard_count >= 2:
                    break
                pid = asteroid_systems[sys_id][0]
                if sys_id in need_yard:
                    pid2 = need_yard[sys_id]
                    res = _try_enqueue(Shipyard.BASE, pid2, at_front=True)
                    if res > 0:
                        queued_shipyard_pids.append(pid2)
                        building_expense += res
                if pid not in queued_building_pids and building_type.can_be_produced(pid):
                    res = _try_enqueue(building_type, pid, at_front=True)
                    if res > 0:
                        new_yard_count += 1
                        queued_building_pids.append(pid)
                        building_expense += res
    return building_expense


def _build_orbital_drydock(top_pilot_systems: TopPilotSystems) -> None:  # noqa: C901
    """Consider building orbital drydocks."""
    building_type = Shipyard.ORBITAL_DRYDOCK
    if building_type.available():
        empire = fo.getEmpire()
        universe = fo.getUniverse()
        queued_pids = building_type.queued_at()
        current_drydock_sys = building_type.built_or_queued_at_sys()
        covered_drydock_systems = set()
        for start_set, dest_set in [
            (current_drydock_sys, covered_drydock_systems),
            (covered_drydock_systems, covered_drydock_systems),
        ]:  # coverage of neighbors up to 2 jumps away from a drydock
            for dd_sys_id in start_set.copy():
                dest_set.add(dd_sys_id)
                dest_set.update(get_neighbors(dd_sys_id))

        max_dock_builds = int(0.8 + empire.productionPoints / 120.0)
        debug(
            "Considering building %s, found current and queued systems %s",
            building_type,
            PlanetUtilsAI.sys_name_ids(current_drydock_sys),
        )
        for sys_id, pids in get_colonized_planets().items():  # TODO: sort/prioritize in some fashion
            local_top_pilots = dict(top_pilot_systems.get(sys_id, []))
            local_drydocks = get_empire_drydocks().get(sys_id, [])
            if len(queued_pids) >= max_dock_builds:
                debug("Drydock enqueing halted with %d of max %d", len(queued_pids), max_dock_builds)
                break
            if (sys_id in covered_drydock_systems) and not local_top_pilots:
                continue
            else:
                pass
            for _, pid in sorted([(local_top_pilots.get(pid, 0), pid) for pid in pids], reverse=True):
                if has_shipyard(pid):
                    continue
                if pid in local_drydocks or pid in queued_pids:
                    break
                if not building_type.can_be_enqueued(pid):
                    continue
                res = _try_enqueue(building_type, pid, at_front=(max_dock_builds >= 2))
                if res > 0:
                    queued_pids.append(pid)
                    system_id = universe.getPlanet(pid).systemID
                    covered_drydock_systems.add(system_id)
                    covered_drydock_systems.update(get_neighbors(system_id))


def _remove_other_colonies(pid: PlanetId, building_name: str) -> None:
    """
    Removes enqueued colony buildings at the given planet.
    Since colonies cannot be queued in parallel, to allow enqueuing building_name, all others must be removed.
    If building_name is already enqueued, it's fine of course.
    """
    numbered_queue = list(enumerate(fo.getEmpire().productionQueue))
    # It should not be more than one, except possibly when loading an old safe file, just to be sure, remove all.
    # Start at the end to avoid changing the numbers of further elements when removing one.
    for num, entry in reversed(numbered_queue):
        if entry.locationID == pid and entry.name.startswith("BLD_COL_") and entry.name != building_name:
            fo.issueDequeueProductionOrder(num)



from stub_generator.interface_inspector.class_processor import ClassInfo
from stub_generator.interface_inspector.enum_processor import EnumInfo
from stub_generator.interface_inspector.function_processor import FunctionInfo
from stub_generator.interface_inspector.instance_processing import InstanceInfo
from stub_generator.interface_inspector.module_inspector import get_module_info



from logging import debug

from common.option_tools import check_bool, get_option_dict

# Minimum evaluation score that a planet must reach to it is considered for outposting or colonizing.
MINIMUM_COLONY_SCORE = 60
# Resource production is evaluated as production * priority_for_the_resource * RESOURCE_PRIORITY_MULTIPLIER.
# So raising this makes it easier for planets to reach the minimum score and in relation reduces the value of
# other scores, which are rated as factor * MINIMUM_COLONY_SCORE
RESOURCE_PRIORITY_MULTIPLIER = 0.5

DEBUG_COLONY_RATING = check_bool(get_option_dict().get("debug_planet_rating", "False"))


def debug_rating(*args, **kwargs):
    if DEBUG_COLONY_RATING:
        debug(*args, **kwargs)



from focs._tech import *
from macros.base_prod import TECH_COST_MULTIPLIER

Tech(
    name="SHP_FIGHTERS_1",
    description="SHP_FIGHTERS_1_DESC",
    short_description="SHIP_WEAPON_UNLOCK_SHORT_DESC",
    category="SHIP_WEAPONS_CATEGORY",
    researchcost=12 * TECH_COST_MULTIPLIER,
    researchturns=3,
    tags=["PEDIA_FIGHTER_TECHS"],
    prerequisites=["SHP_ROOT_AGGRESSION"],
    unlock=[
        Item(type=UnlockShipPart, name="FT_BAY_1"),
        Item(type=UnlockShipPart, name="FT_HANGAR_1"),
        Item(type=UnlockShipPart, name="FT_HANGAR_2"),
        Item(type=UnlockShipPart, name="FT_HANGAR_3"),
        Item(type=UnlockShipPart, name="FT_HANGAR_4"),
    ],
    graphic="icons/ship_parts/fighter05.png",
)



# Main ordered stages:
# 1. Effects with SetTargetPopulation
# 2. Effects with SetPopulation
# 3. Effects with SetTarget<other meters>
# 4. Effects with Set<other meters>
# 5. Final stage (max priority overrides and clean ups)

# Default priority: 100 = DEFAULT_PRIORITY = TARGET_BEFORE_SCALING_PRIORITY

# First stage: Target Population

BEFORE_ANYTHING_ELSE_PRIORITY = 1

TARGET_POPULATION_BEFORE_SCALING_PRIORITY = 13

TARGET_POPULATION_SCALING_PRIORITY = 15

TARGET_POPULATION_AFTER_SCALING_PRIORITY = 17

TARGET_POPULATION_2ND_SCALING_PRIORITY = 20

TARGET_POPULATION_AFTER_2ND_SCALING_PRIORITY = 22

TARGET_POPULATION_LAST_BEFORE_OVERRIDE_PRIORITY = 29

TARGET_POPULATION_OVERRIDE_PRIORITY = 30


# Second stage: Population

POPULATION_FIRST_PRIORITY = 40

POPULATION_DEFAULT_PRIORITY = 45

POPULATION_OVERRIDE_PRIORITY = 50


# Third stage: Other target meters

TARGET_FIRST_PRIORITY = 95

TARGET_EARLY_BEFORE_SCALING_PRIORITY = 98

DEFAULT_PRIORITY = 100

TARGET_BEFORE_SCALING_PRIORITY = DEFAULT_PRIORITY

TARGET_LATE_BEFORE_SCALING_PRIORITY = 102

TARGET_SCALING_PRIORITY = 105

TARGET_EARLY_AFTER_SCALING_PRIORITY = 108

TARGET_AFTER_SCALING_PRIORITY = 110

TARGET_LATE_AFTER_SCALING_PRIORITY = 112

TARGET_2ND_SCALING_PRIORITY = 115

TARGET_EARLY_AFTER_2ND_SCALING_PRIORITY = 118

TARGET_AFTER_2ND_SCALING_PRIORITY = 120

TARGET_LATE_AFTER_2ND_SCALING_PRIORITY = 122

TARGET_LAST_BEFORE_OVERRIDE_PRIORITY = 199

TARGET_OVERRIDE_PRIORITY = 200


# Fourth stage: current meters

EARLY_AFTER_ALL_TARGET_MAX_METERS_PRIORITY = 490

AFTER_ALL_TARGET_MAX_METERS_PRIORITY = 500

LATE_AFTER_ALL_TARGET_MAX_METERS_PRIORITY = 510


# Fifth stage: max priority overrides and cleanups

FOCUS_CHANGE_PENALTY_PRIORITY = 600

METER_OVERRIDE_PRIORITY = 700

CONCENTRATION_CAMP_PRIORITY = 800

SPECIAL_REMOVAL_PRIORITY = 900

END_CLEANUP_PRIORITY = 1000



from focs._effects import (
    DesignHasPart,
    EffectsGroup,
    OwnedBy,
    PartsInShipDesign,
    SetMaxCapacity,
    Ship,
    Source,
    Target,
    Value,
)
from focs._tech import *
from macros.base_prod import TECH_COST_MULTIPLIER
from techs.ship_weapons.ship_weapons import HANGAR_UPGRADE_SECONDARY_STAT_EFFECT

Tech(
    name="SHP_FIGHTERS_3",
    description="SHP_FIGHTERS_3_DESC",
    short_description="SHIP_WEAPON_IMPROVE_SHORT_DESC",
    category="SHIP_WEAPONS_CATEGORY",
    researchcost=360 * TECH_COST_MULTIPLIER,
    researchturns=9,
    tags=["PEDIA_FIGHTER_TECHS"],
    prerequisites=["SHP_FIGHTERS_2"],
    effectsgroups=[
        EffectsGroup(
            scope=Ship
            & OwnedBy(empire=Source.Owner)
            & (
                DesignHasPart(name="FT_HANGAR_1")
                | DesignHasPart(name="FT_HANGAR_2")
                | DesignHasPart(name="FT_HANGAR_3")
                | DesignHasPart(name="FT_HANGAR_4")
            ),
            accountinglabel="SHP_FIGHTERS_3",
            effects=[
                SetMaxCapacity(
                    partname="FT_HANGAR_1", value=Value + PartsInShipDesign(name="FT_HANGAR_1", design=Target.DesignID)
                ),
                HANGAR_UPGRADE_SECONDARY_STAT_EFFECT("SHP_FIGHTERS_3", "FT_HANGAR_2", 2),
                HANGAR_UPGRADE_SECONDARY_STAT_EFFECT("SHP_FIGHTERS_3", "FT_HANGAR_3", 3),
                HANGAR_UPGRADE_SECONDARY_STAT_EFFECT("SHP_FIGHTERS_3", "FT_HANGAR_4", 6),
            ],
        )
    ],
    graphic="icons/ship_parts/fighter05.png",
)



from logging import info, warning

from common.configure_logging import redirect_logging_to_freeorion_logger

# Logging is redirected before other imports so that import errors appear in log files.
redirect_logging_to_freeorion_logger()

import freeorion as fo
import sys

# Constants defined by the C++ game engine
NO_TEAM_ID = -1


class AuthProvider:
    def __init__(self):
        self.logins = {}
        self.roles_symbols = {
            "h": fo.roleType.host,
            "m": fo.roleType.clientTypeModerator,
            "p": fo.roleType.clientTypePlayer,
            "o": fo.roleType.clientTypeObserver,
            "g": fo.roleType.galaxySetup,
        }
        try:
            with open(fo.get_user_config_dir() + "/auth.txt") as f:
                first_line = True
                for line in f:
                    if first_line:
                        first_line = False
                        self.default_roles = self.__parse_roles(line.strip())
                    else:
                        login, roles, password = line.rsplit(":", 2)
                        self.logins[login] = (password.strip(), self.__parse_roles(roles.strip()))
        except OSError:
            exctype, value = sys.exc_info()[:2]
            warning("Can't read auth file {}: {} {}".format(fo.get_user_config_dir() + "/auth.txt", exctype, value))
            self.default_roles = [
                fo.roleType.clientTypeModerator,
                fo.roleType.clientTypePlayer,
                fo.roleType.clientTypeObserver,
                fo.roleType.galaxySetup,
            ]
        info("Auth initialized")

    def __parse_roles(self, roles_str):
        roles = []
        for c in roles_str:
            r = self.roles_symbols.get(c)
            if r is None:
                warning("unknown role symbol '%c'" % c)
            else:
                roles.append(r)
        return roles

    def is_require_auth_or_return_roles(self, player_name: str, ip_address: str):
        """Returns True if player should be authenticated, False if user not allowed,
        or list of roles for anonymous players"""
        known_login = player_name in self.logins
        if not known_login:
            # default list of roles
            return self.default_roles
        return True

    def is_success_auth_and_return_roles(self, player_name, auth):
        """Return False if passowrd doesn't match or list of roles for authenticated player"""
        auth_data = self.logins.get(player_name)
        if auth_data[0] == auth:
            return auth_data[1]
        else:
            return False

    def list_players(self):
        """Returns list of PlayerSetupData to use in quickstart"""
        players = []
        for player_name, auth_data in self.logins.items():
            if fo.roleType.clientTypePlayer in auth_data[1]:
                psd = fo.PlayerSetupData()
                psd.player_name = player_name
                psd.empire_name = player_name
                psd.starting_species = "RANDOM"
                psd.starting_team = NO_TEAM_ID
                players.append(psd)
        return players

    def get_player_delegation(self, player_name):
        """Returns list of players delegated by this player"""
        return []



import freeOrionAIInterface as fo
import math
from collections.abc import Sequence
from copy import copy
from logging import debug, info, warning

import AIDependencies
import AIstate
import ColonisationAI
import EspionageAI
import FleetUtilsAI
import MilitaryAI
import PlanetUtilsAI
from AIDependencies import INVALID_ID
from aistate_interface import get_aistate
from colonization import calculate_planet_colonization_rating
from common.fo_typing import SystemId
from common.print_utils import Number, Table, Text
from DiplomaticCorp import get_diplomatic_status
from empire.ship_builders import can_build_ship_for_species
from EnumsAI import EmpireProductionTypes, MissionType, PriorityType, ShipRoleType
from freeorion_tools import (
    get_partial_visibility_turn,
    get_species_attack_troops,
    tech_is_complete,
)
from freeorion_tools.caching import cache_for_current_turn
from freeorion_tools.timers import AITimer
from target import TargetPlanet, TargetSystem
from turn_state import get_colonized_planets_in_system
from turn_state.design import get_best_ship_info
from universe.system_network import systems_connected

MAX_BASE_TROOPERS_GOOD_INVADERS = 20
MAX_BASE_TROOPERS_POOR_INVADERS = 10
_TROOPS_SAFETY_MARGIN = 1  # try to send this amount of additional troops to account for uncertainties in calculation
MIN_INVASION_SCORE = 20

invasion_timer = AITimer("get_invasion_fleets()")


def get_invasion_fleets():  # noqa: C901
    invasion_timer.start("gathering initial info")
    universe = fo.getUniverse()
    empire = fo.getEmpire()
    empire_id = fo.empireID()

    home_system_id = PlanetUtilsAI.get_capital_sys_id()
    aistate = get_aistate()
    visible_system_ids = list(aistate.visInteriorSystemIDs) + list(aistate.visBorderSystemIDs)

    if home_system_id != INVALID_ID:
        accessible_system_ids = [sys_id for sys_id in visible_system_ids if systems_connected(sys_id, home_system_id)]
    else:
        debug("Empire has no identifiable homeworld; will treat all visible planets as accessible.")
        # TODO: check if any troop ships owned, use their system as home system
        accessible_system_ids = visible_system_ids

    acessible_planet_ids = PlanetUtilsAI.get_planets_in__systems_ids(accessible_system_ids)
    all_owned_planet_ids = PlanetUtilsAI.get_all_owned_planet_ids(acessible_planet_ids)  # includes unpopulated outposts
    all_populated_planets = PlanetUtilsAI.get_populated_planet_ids(acessible_planet_ids)  # includes unowned natives
    empire_owned_planet_ids = PlanetUtilsAI.get_owned_planets_by_empire()
    invadable_planet_ids = set(all_owned_planet_ids).union(all_populated_planets) - set(empire_owned_planet_ids)

    invasion_targeted_planet_ids = get_invasion_targeted_planet_ids(universe.planetIDs, MissionType.INVASION)
    invasion_targeted_planet_ids.extend(
        get_invasion_targeted_planet_ids(universe.planetIDs, MissionType.ORBITAL_INVASION)
    )
    all_invasion_targeted_system_ids = set(PlanetUtilsAI.get_systems(invasion_targeted_planet_ids))

    invasion_fleet_ids = FleetUtilsAI.get_empire_fleet_ids_by_role(MissionType.INVASION)
    num_invasion_fleets = len(FleetUtilsAI.extract_fleet_ids_without_mission_types(invasion_fleet_ids))

    debug("Current Invasion Targeted SystemIDs: %s" % PlanetUtilsAI.sys_name_ids(AIstate.invasionTargetedSystemIDs))
    debug("Current Invasion Targeted PlanetIDs: %s" % PlanetUtilsAI.planet_string(invasion_targeted_planet_ids))
    debug(invasion_fleet_ids and "Invasion Fleet IDs: %s" % invasion_fleet_ids or "Available Invasion Fleets: 0")
    debug("Invasion Fleets Without Missions: %s" % num_invasion_fleets)

    invasion_timer.start("planning troop base production")
    reserved_troop_base_targets = []
    if aistate.character.may_invade_with_bases():
        available_pp = {}
        for el in empire.planetsWithAvailablePP:  # keys are sets of ints; data is doubles
            avail_pp = el.data()
            for pid in el.key():
                available_pp[pid] = avail_pp
        # For planning base trooper invasion targets we have a two-pass system.  (1) In the first pass we consider all
        # the invasion targets and figure out which ones appear to be suitable for using base troopers against (i.e., we
        # already have a populated planet in the same system that could build base troopers) and we have at least a
        # minimal amount of PP available, and (2) in the second pass we go through the reserved base trooper target list
        # and check to make sure that there does not appear to be too much military action still needed before the
        # target is ready to be invaded, we double check that not too many base troopers would be needed, and if things
        # look clear then we queue up the base troopers on the Production Queue and keep track of where we are building
        # them, and how many; we may also disqualify and remove previously qualified targets (in case, for example,
        # we lost our base trooper source planet since it was first added to list).
        #
        # For planning and tracking base troopers under construction, we use a dictionary store in
        # get_aistate().qualifyingTroopBaseTargets, keyed by the invasion target planet ID.  We only store values
        # for invasion targets that appear likely to be suitable for base trooper use, and store a 2-item list.
        # The first item in this list is the ID of the planet where we expect to build the base troopers, and the second
        # entry initially is set to INVALID_ID (-1).  The presence of this entry in qualifyingTroopBaseTargets
        # flags this target as being reserved as a base-trooper invasion target.
        # In the second pass, if/when we actually start construction, then we modify the record, replacing that second
        # value with the ID of the planet where the troopers are actually being built.  (Right now that is always the
        # same as the source planet originally identified, but we could consider reevaluating that, or use that second
        # value to instead record how many base troopers have been queued, so that on later turns we can assess if the
        # process got delayed & perhaps more troopers need to be queued).

        # Pass 1: identify qualifying base troop invasion targets
        for pid in invadable_planet_ids:  # TODO: reorganize
            if pid in aistate.qualifyingTroopBaseTargets:
                continue
            planet = universe.getPlanet(pid)
            if not planet:
                continue
            sys_id = planet.systemID
            sys_partial_vis_turn = get_partial_visibility_turn(sys_id)
            planet_partial_vis_turn = get_partial_visibility_turn(pid)
            if planet_partial_vis_turn < sys_partial_vis_turn:
                continue
            best_base_planet = INVALID_ID
            best_trooper_count = 0
            for pid2 in get_colonized_planets_in_system(sys_id):
                if available_pp.get(pid2, 0) < 2:  # TODO: improve troop base PP sufficiency determination
                    break
                planet2 = universe.getPlanet(pid2)
                if not planet2 or not can_build_ship_for_species(planet2.speciesName):
                    continue
                best_base_trooper_here = get_best_ship_info(PriorityType.PRODUCTION_ORBITAL_INVASION, pid2)[1]
                if not best_base_trooper_here:
                    continue
                troops_per_ship = best_base_trooper_here.troopCapacity
                if not troops_per_ship:
                    continue
                troops_per_ship = troops_per_ship * get_species_attack_troops(planet2.speciesName)
                if troops_per_ship > best_trooper_count:
                    best_base_planet = pid2
                    best_trooper_count = troops_per_ship
            if best_base_planet != INVALID_ID:
                aistate.qualifyingTroopBaseTargets.setdefault(pid, [best_base_planet, INVALID_ID])

        # Pass 2: for each target previously identified for base troopers, check that still qualifies and
        # check how many base troopers would be needed; if reasonable then queue up the troops and record this in
        # get_aistate().qualifyingTroopBaseTargets
        for pid in list(aistate.qualifyingTroopBaseTargets.keys()):
            planet = universe.getPlanet(pid)
            if planet and planet.owner == empire_id:
                del aistate.qualifyingTroopBaseTargets[pid]
                continue
            if pid in invasion_targeted_planet_ids:  # TODO: consider overriding standard invasion mission
                continue
            if aistate.qualifyingTroopBaseTargets[pid][1] != -1:
                reserved_troop_base_targets.append(pid)
                if planet:
                    all_invasion_targeted_system_ids.add(planet.systemID)
                # TODO: evaluate changes to situation, any more troops needed, etc.
                continue  # already building for here
            _, planet_troops = evaluate_invasion_planet(pid)
            sys_id = planet.systemID
            this_sys_status = aistate.systemStatus.get(sys_id, {})
            troop_tally = _get_base_troopers(sys_id)
            if troop_tally > planet_troops:  # base troopers appear unneeded
                del aistate.qualifyingTroopBaseTargets[pid]
                continue
            planet_troops -= troop_tally
            if planet.currentMeterValue(fo.meterType.shield) > 0 and (
                this_sys_status.get("myFleetRating", 0) < 0.8 * this_sys_status.get("totalThreat", 0)
                or this_sys_status.get("myFleetRatingVsPlanets", 0) < this_sys_status.get("planetThreat", 0)
            ):
                # this system not secured, so ruling out invasion base troops for now
                # don't immediately delete from qualifyingTroopBaseTargets or it will be opened up for regular troops
                continue
            loc = aistate.qualifyingTroopBaseTargets[pid][0]
            best_base_trooper_here = get_best_ship_info(PriorityType.PRODUCTION_ORBITAL_INVASION, loc)[1]
            loc_planet = universe.getPlanet(loc)
            if best_base_trooper_here is None:  # shouldn't be possible at this point, but just to be safe
                warning("Could not find a suitable orbital invasion design at %s" % loc_planet)
                continue
            # TODO: have TroopShipDesigner give the expected number of troops including species effects directly
            troops_per_ship = best_base_trooper_here.troopCapacity
            troops_per_ship = troops_per_ship * get_species_attack_troops(loc_planet.speciesName)
            if not troops_per_ship:
                warning("The best orbital invasion design at %s seems not to have any troop capacity." % loc_planet)
                continue
            _, col_design, build_choices = get_best_ship_info(PriorityType.PRODUCTION_ORBITAL_INVASION, loc)
            if not col_design:
                continue
            if loc not in build_choices:
                warning(f"Best troop design {col_design} can not be produced at planet with id: {build_choices}")
                continue
            n_bases = math.ceil((planet_troops + 1) / troops_per_ship)  # TODO: reconsider this +1 safety factor
            # TODO: evaluate cost and time-to-build of best base trooper here versus cost and time-to-build-and-travel
            # for best regular trooper elsewhere
            # For now, we assume what building base troopers is best so long as either (1) we would need no more than
            # MAX_BASE_TROOPERS_POOR_INVADERS base troop ships, or (2) our base troopers have more than 1 trooper per
            # ship and we would need no more than MAX_BASE_TROOPERS_GOOD_INVADERS base troop ships
            if n_bases > MAX_BASE_TROOPERS_POOR_INVADERS or (
                troops_per_ship > 1 and n_bases > MAX_BASE_TROOPERS_GOOD_INVADERS
            ):
                debug("ruling out base invasion troopers for %s due to high number (%d) required." % (planet, n_bases))
                del aistate.qualifyingTroopBaseTargets[pid]
                continue
            debug(
                "Invasion base planning, need %d troops at %d per ship, will build %d ships."
                % ((planet_troops + 1), troops_per_ship, n_bases)
            )
            retval = fo.issueEnqueueShipProductionOrder(col_design.id, loc)
            debug(
                "Enqueueing %d Troop Bases at %s for %s"
                % (n_bases, PlanetUtilsAI.planet_string(loc), PlanetUtilsAI.planet_string(pid))
            )
            if retval != 0:
                all_invasion_targeted_system_ids.add(planet.systemID)
                reserved_troop_base_targets.append(pid)
                aistate.qualifyingTroopBaseTargets[pid][1] = loc
                fo.issueChangeProductionQuantityOrder(empire.productionQueue.size - 1, 1, int(n_bases))
                fo.issueRequeueProductionOrder(empire.productionQueue.size - 1, 0)

    invasion_timer.start("evaluating target planets")
    # TODO: check if any invasion_targeted_planet_ids need more troops assigned
    evaluated_planet_ids = list(
        set(invadable_planet_ids) - set(invasion_targeted_planet_ids) - set(reserved_troop_base_targets)
    )
    evaluated_planets = assign_invasion_values(evaluated_planet_ids)

    sorted_planets = [(pid, pscore % 10000, ptroops) for pid, (pscore, ptroops) in evaluated_planets.items()]
    sorted_planets.sort(key=lambda x: x[1], reverse=True)
    sorted_planets = [(pid, pscore % 10000, ptroops) for pid, pscore, ptroops in sorted_planets]

    invasion_table = Table(
        Text("Planet"),
        Number("Score"),
        Text("Species"),
        Number("Troops"),
        table_name="Potential Targets for Invasion Turn %d" % fo.currentTurn(),
    )

    for pid, pscore, ptroops in sorted_planets:
        planet = universe.getPlanet(pid)
        invasion_table.add_row(planet, pscore, planet and planet.speciesName or "unknown", ptroops)
    info(invasion_table)

    sorted_planets = [x for x in sorted_planets if x[1] > 0]
    # export opponent planets for other AI modules
    AIstate.opponentPlanetIDs = [pid for pid, __, __ in sorted_planets]
    AIstate.invasionTargets = sorted_planets

    # export invasion targeted systems for other AI modules
    AIstate.invasionTargetedSystemIDs = list(all_invasion_targeted_system_ids)
    invasion_timer.stop(section_name="evaluating %d target planets" % (len(evaluated_planet_ids)))
    invasion_timer.stop_print_and_clear()


def get_invasion_targeted_planet_ids(planet_ids, mission_type):
    invasion_feet_missions = get_aistate().get_fleet_missions_with_any_mission_types([mission_type])
    targeted_planets = []
    for pid in planet_ids:
        # add planets that are target of a mission
        for mission in invasion_feet_missions:
            target = TargetPlanet(pid)
            if mission.has_target(mission_type, target):
                targeted_planets.append(pid)
    return targeted_planets


def retaliation_risk_factor(empire_id):
    """A multiplicative adjustment to planet scores to account for risk of retaliation from planet owner."""
    # TODO implement (in militaryAI) actual military risk assessment of other empires
    if empire_id == -1:  # unowned
        return 1.5  # since no risk of retaliation, increase score
    else:
        return 1.0


def assign_invasion_values(planet_ids):
    """Creates a dictionary that takes planet_ids as key and their invasion score as value."""
    empire_id = fo.empireID()
    planet_values = {}
    neighbor_values = {}
    neighbor_val_ratio = 0.3
    universe = fo.getUniverse()
    for pid in planet_ids:
        # Must copy to avoid modifying neighbor_values when adding neighbor_values to planet_values
        planet_values[pid] = copy(neighbor_values.setdefault(pid, evaluate_invasion_planet(pid)))
        debug("planet %d, values %s", pid, planet_values[pid])
        planet = universe.getPlanet(pid)
        species_name = (planet and planet.speciesName) or ""
        species = fo.getSpecies(species_name)
        if species and species.canProduceShips:
            system = universe.getSystem(planet.systemID)
            if not system:
                continue
            planet_industries = {}
            for pid2 in system.planetIDs:
                planet2 = universe.getPlanet(pid2)
                species_name2 = (planet2 and planet2.speciesName) or ""
                species2 = fo.getSpecies(species_name2)
                if species2 and species2.canProduceShips:
                    # to prevent divide-by-zero
                    planet_industries[pid2] = planet2.initialMeterValue(fo.meterType.industry) + 0.1
            industry_ratio = planet_industries[pid] / max(planet_industries.values())
            for pid2 in system.planetIDs:
                if pid2 == pid:
                    continue
                planet2 = universe.getPlanet(pid2)
                # TODO check for allies
                if (
                    planet2
                    and (planet2.owner != empire_id)
                    and ((planet2.owner != -1) or (planet2.initialMeterValue(fo.meterType.population) > 0))
                ):
                    planet_values[pid][0] += (
                        industry_ratio
                        * neighbor_val_ratio
                        * (neighbor_values.setdefault(pid2, evaluate_invasion_planet(pid2))[0])
                    )
    return planet_values


@cache_for_current_turn
def _get_path_from_capital(planet: "fo.planet") -> tuple[Sequence[SystemId], int]:
    """
    Return chain of system from the planet to capital and length.

    If there is no path, return empty sequence and default distance.
    """
    universe = fo.getUniverse()
    capital_id = PlanetUtilsAI.get_capital()
    if capital_id != INVALID_ID:
        homeworld = universe.getPlanet(capital_id)
        if homeworld and homeworld.systemID != INVALID_ID and planet.systemID != INVALID_ID:
            least_jumps_path = list(universe.leastJumpsPath(homeworld.systemID, planet.systemID, fo.empireID()))
            max_jumps = len(least_jumps_path)
            return least_jumps_path, max_jumps
    else:
        return [], 8


def evaluate_invasion_planet(planet_id):  # noqa: C901
    """Return the invasion value (score, troops) of a planet."""
    universe = fo.getUniverse()
    empire_id = fo.empireID()
    detail = []

    planet = universe.getPlanet(planet_id)
    if planet is None:
        debug("Invasion AI couldn't access any info for planet id %d" % planet_id)
        return [0, 0]

    if planet.owner != INVALID_ID and get_diplomatic_status(planet.owner) != fo.diplomaticStatus.war:
        return [0, 0]

    system_id = planet.systemID

    # by using the following instead of simply relying on stealth meter reading,
    # can (sometimes) plan ahead even if planet is temporarily shrouded by an ion storm
    predicted_detectable = EspionageAI.colony_detectable_by_empire(
        planet_id, empire=fo.empireID(), default_result=False
    )
    if not predicted_detectable:
        if get_partial_visibility_turn(planet_id) < fo.currentTurn():
            debug("InvasionAI predicts planet id %d to be stealthed" % planet_id)
            return [0, 0]
        else:
            debug(
                "InvasionAI predicts planet id %d to be stealthed" % planet_id
                + ", but somehow have current visibility anyway, will still consider as target"
            )

    # Check if the target planet was extra-stealthed somehow its system was last viewed
    # this test below may augment the tests above,
    # but can be thrown off by temporary combat-related sighting
    system_last_seen = get_partial_visibility_turn(planet_id)
    planet_last_seen = get_partial_visibility_turn(system_id)
    if planet_last_seen < system_last_seen:
        # TODO: track detection strength, order new scouting when it goes up
        debug("Invasion AI considering planet id %d (stealthed at last view), still proceeding." % planet_id)

    # get a baseline evaluation of the planet as determined by ColonisationAI
    species_name = planet.speciesName
    species = fo.getSpecies(species_name)
    if not species or AIDependencies.TAG_DESTROYED_ON_CONQUEST in species.tags:
        # this call iterates over this Empire's available species with which it could colonize after an invasion
        planet_eval = ColonisationAI.assign_colonisation_values([planet_id], MissionType.INVASION, None, detail)
        colony_base_value = max(
            0.75 * planet_eval.get(planet_id, [0])[0],
            calculate_planet_colonization_rating.calculate_planet_colonization_rating(
                planet_id=planet_id,
                mission_type=MissionType.OUTPOST,
                spec_name=None,
                detail=detail,
            ),
        )
    else:
        colony_base_value = calculate_planet_colonization_rating.calculate_planet_colonization_rating(
            planet_id=planet_id,
            mission_type=MissionType.INVASION,
            spec_name=species_name,
            detail=detail,
        )

    # Add extra score for all buildings on the planet
    building_values = {
        "BLD_IMPERIAL_PALACE": 1000,
        "BLD_CULTURE_ARCHIVES": 1000,
        "BLD_AUTO_HISTORY_ANALYSER": 100,
        "BLD_SHIPYARD_BASE": 100,
        "BLD_SHIPYARD_ORG_ORB_INC": 200,
        "BLD_SHIPYARD_ORG_XENO_FAC": 200,
        "BLD_SHIPYARD_ORG_CELL_GRO_CHAMB": 200,
        "BLD_SHIPYARD_CON_NANOROBO": 300,
        "BLD_SHIPYARD_CON_GEOINT": 400,
        "BLD_SHIPYARD_CON_ADV_ENGINE": 1000,
        "BLD_SHIPYARD_AST": 300,
        "BLD_SHIPYARD_AST_REF": 1000,
        "BLD_SHIPYARD_ENRG_SOLAR": 1500,
        "BLD_INDUSTRY_CENTER": 500,
        "BLD_GAS_GIANT_GEN": 200,
        "BLD_SOL_ORB_GEN": 800,
        "BLD_BLACK_HOLE_POW_GEN": 2000,
        "BLD_ENCLAVE_VOID": 500,
        "BLD_NEUTRONIUM_EXTRACTOR": 2000,
        "BLD_NEUTRONIUM_SYNTH": 2000,
        "BLD_NEUTRONIUM_FORGE": 1000,
        "BLD_CONC_CAMP": 100,
        "BLD_BIOTERROR_PROJECTOR": 1000,
        "BLD_SHIPYARD_ENRG_COMP": 3000,
    }
    bld_tally = 0
    for bldType in [universe.getBuilding(bldg).buildingTypeName for bldg in planet.buildingIDs]:
        bval = building_values.get(bldType, 50)
        bld_tally += bval
        detail.append("%s: %d" % (bldType, bval))
    # quick fix, building values dwarf the new planet ratings, and I think they are too high in general
    bld_tally /= 20

    # Add extra score for unlocked techs when we conquer the species
    tech_tally = 0
    value_per_pp = 4
    for unlocked_tech in AIDependencies.SPECIES_TECH_UNLOCKS.get(species_name, []):
        if not tech_is_complete(unlocked_tech):
            rp_cost = fo.getTech(unlocked_tech).researchCost(empire_id)
            tech_value = value_per_pp * rp_cost
            tech_tally += tech_value
            detail.append("%s: %d" % (unlocked_tech, tech_value))

    least_jumps_path, max_jumps = _get_path_from_capital(planet)
    clear_path = True

    aistate = get_aistate()
    system_status = aistate.systemStatus.get(system_id, {})
    system_fleet_treat = system_status.get("fleetThreat", 1000)
    system_monster_threat = system_status.get("monsterThreat", 0)
    sys_total_threat = system_fleet_treat + system_monster_threat + system_status.get("planetThreat", 0)
    max_path_threat = system_fleet_treat
    mil_ship_rating = MilitaryAI.cur_best_mil_ship_rating()
    for path_sys_id in least_jumps_path:
        path_leg_status = aistate.systemStatus.get(path_sys_id, {})
        path_leg_threat = path_leg_status.get("fleetThreat", 1000) + path_leg_status.get("monsterThreat", 0)
        if path_leg_threat > 0.5 * mil_ship_rating:
            clear_path = False
            if path_leg_threat > max_path_threat:
                max_path_threat = path_leg_threat

    pop = planet.currentMeterValue(fo.meterType.population)
    target_pop = planet.currentMeterValue(fo.meterType.targetPopulation)
    troops = planet.currentMeterValue(fo.meterType.troops)
    troop_regen = planet.currentMeterValue(fo.meterType.troops) - planet.initialMeterValue(fo.meterType.troops)
    max_troops = planet.currentMeterValue(fo.meterType.maxTroops)
    # TODO: refactor troop determination into function for use in mid-mission updates and also consider defender techs
    max_troops += AIDependencies.TROOPS_PER_POP * (target_pop - pop)
    system_secured = secure_system(system_id, False)
    debug(
        "Invasion eval of %s\n" " - maxShields: %.1f\n" " - sysFleetThreat: %.1f\n" " - sysMonsterThreat: %.1f",
        planet,
        planet.currentMeterValue(fo.meterType.maxShield),
        system_fleet_treat,
        system_monster_threat,
    )
    enemy_val = 0
    if planet.owner != -1:  # value in taking this away from an enemy
        enemy_val = 20 * (
            planet.currentMeterValue(fo.meterType.targetIndustry)
            + 2 * planet.currentMeterValue(fo.meterType.targetResearch)
        )

    # devalue invasions that would require too much military force
    preferred_max_portion = MilitaryAI.get_preferred_max_military_portion_for_single_battle()
    total_max_mil_rating = MilitaryAI.get_concentrated_tot_mil_rating()
    threat_exponent = 2  # TODO: make this a character trait; higher aggression with a lower exponent
    threat_factor = min(1, preferred_max_portion * total_max_mil_rating / (sys_total_threat + 0.001)) ** threat_exponent

    design_id, _, locs = get_best_ship_info(PriorityType.PRODUCTION_INVASION)
    if not locs or not universe.getPlanet(locs[0]):
        # We are in trouble anyway, so just calculate whatever approximation...
        build_time = 4
        planned_troops = troops if system_secured else min(troops + troop_regen * (max_jumps + build_time), max_troops)
        planned_troops += 0.01  # we must attack with more troops than there are defenders
        troop_cost = math.ceil((planned_troops + _TROOPS_SAFETY_MARGIN) / 6.0) * 20 * FleetUtilsAI.get_fleet_upkeep()
    else:
        loc = locs[0]
        species_here = universe.getPlanet(loc).speciesName
        design = fo.getShipDesign(design_id)
        cost_per_ship = design.productionCost(empire_id, loc)
        build_time = design.productionTime(empire_id, loc)
        troops_per_ship = design.troopCapacity * get_species_attack_troops(species_here)
        planned_troops = troops if system_secured else min(troops + troop_regen * (max_jumps + build_time), max_troops)
        planned_troops += 0.01  # we must attack with more troops than there are defenders
        ships_needed = math.ceil((planned_troops + _TROOPS_SAFETY_MARGIN) / float(troops_per_ship))
        troop_cost = ships_needed * cost_per_ship  # fleet upkeep is already included in query from server

    # apply some bias to expensive operations
    normalized_cost = float(troop_cost) / max(fo.getEmpire().productionPoints, 1)
    normalized_cost = max(1.0, normalized_cost)
    cost_score = (normalized_cost**2 / 50.0) * troop_cost

    base_score = colony_base_value + bld_tally + tech_tally + enemy_val - cost_score
    # If the AI does have enough total military to attack this target, and the target is more than minimally valuable,
    # don't let the threat_factor discount the adjusted value below MIN_INVASION_SCORE +1, so that if there are no
    # other targets the AI could still pursue this one.  Otherwise, scoring pressure from
    # MilitaryAI.get_preferred_max_military_portion_for_single_battle might prevent the AI from attacking heavily
    # defended but still defeatable targets even if it has no softer targets available.
    if total_max_mil_rating > sys_total_threat and base_score > 2 * MIN_INVASION_SCORE:
        threat_factor = max(threat_factor, (MIN_INVASION_SCORE + 1) / base_score)
    planet_score = retaliation_risk_factor(planet.owner) * threat_factor * max(0, base_score)
    if clear_path:
        planet_score *= 1.5
    debug(
        f" - planet score: {planet_score:.2f}\n"
        f" - planned troops: {planned_troops:.2f}\n"
        f" - projected troop cost: {troop_cost:.1f}\n"
        f" - threat factor: {threat_factor}\n"
        f" - planet detail: {detail}\n"
        f" - popval: {colony_base_value:.1f}\n"
        f" - bldval: {bld_tally}\n"
        f" - enemyval: {enemy_val}\n"
        f" - system secured: {system_secured}"
    )
    return [planet_score, planned_troops]


def secure_system(system_id: SystemId, secure_now: bool) -> bool:
    """Check whether we have any military fleets in the system.
    If secure_now is True, set one of the fleets to MissionType.SECURE
    If secure_now is False, return True if there is any fleet that could secure it.
    """
    universe = fo.getUniverse()
    ai_state = get_aistate()
    system = universe.getSystem(system_id)
    secure_targets = [system_id] + list(system.planetIDs)
    system_status = ai_state.systemStatus.get(system_id, {})
    mission_types = [MissionType.SECURE, MissionType.MILITARY, MissionType.PROTECT_REGION, MissionType.ORBITAL_DEFENSE]
    missions = ai_state.get_fleet_missions_with_any_mission_types(mission_types)
    secured = False
    for mission in missions:
        fleet = universe.getFleet(mission.fleet.id)
        if not fleet or fleet.systemID != system_id:
            continue
        target_obj = mission.target.get_object()
        if target_obj is not None and target_obj.id in secure_targets:
            if secure_now and mission.type != MissionType.SECURE:
                mission.set_target(MissionType.SECURE, mission.target)
            # TBD whether fleet itself has a sufficient rating against a planet.
            # E.g. a fleet with Flux Lance cannot keep planetary shields down.
            secured = system_status.get("myFleetRating", 0) > 0
            break
    return secured


def send_invasion_fleets(fleet_ids, evaluated_planets, mission_type):
    """sends a list of invasion fleets to a list of planet_value_pairs"""
    if not fleet_ids:
        return

    universe = fo.getUniverse()
    invasion_fleet_pool = set(fleet_ids)

    for planet_id, pscore, ptroops in evaluated_planets:
        if pscore < MIN_INVASION_SCORE:
            continue
        planet = universe.getPlanet(planet_id)
        if not planet:
            continue
        sys_id = planet.systemID
        found_fleets = []
        found_stats = {}
        min_stats = {"troopCapacity": ptroops}
        target_stats = {
            "troopCapacity": ptroops + _TROOPS_SAFETY_MARGIN,
            "target_system": TargetSystem(sys_id),
        }
        these_fleets = FleetUtilsAI.get_fleets_for_mission(
            target_stats,
            min_stats,
            found_stats,
            starting_system=sys_id,
            fleet_pool_set=invasion_fleet_pool,
            fleet_list=found_fleets,
        )
        if not these_fleets:
            if not FleetUtilsAI.stats_meet_reqs(found_stats, min_stats):
                debug(
                    "Insufficient invasion troop allocation for system %d ( %s ) -- requested %s , found %s"
                    % (sys_id, universe.getSystem(sys_id).name, min_stats, found_stats)
                )
                invasion_fleet_pool.update(found_fleets)
                continue
            else:
                these_fleets = found_fleets
        target = TargetPlanet(planet_id)
        debug(f"assigning invasion fleets {these_fleets} to target {target}")
        if target.get_object().currentMeterValue(fo.meterType.maxShield) > 0.0 and not secure_system(sys_id, True):
            continue
        aistate = get_aistate()
        for fleetID in these_fleets:
            fleet_mission = aistate.get_fleet_mission(fleetID)
            fleet_mission.clear_fleet_orders()
            fleet_mission.clear_target()
            fleet_mission.set_target(mission_type, target)
        # cannot wait for next turn, else the secure mission may be abandoned
        if these_fleets and sys_id not in AIstate.invasionTargetedSystemIDs:
            AIstate.invasionTargetedSystemIDs.append(sys_id)


def assign_invasion_bases():
    """Assign our troop bases to invasion targets."""
    universe = fo.getUniverse()
    all_troopbase_fleet_ids = FleetUtilsAI.get_empire_fleet_ids_by_role(MissionType.ORBITAL_INVASION)
    available_troopbase_fleet_ids = set(FleetUtilsAI.extract_fleet_ids_without_mission_types(all_troopbase_fleet_ids))

    aistate = get_aistate()
    for fid in list(available_troopbase_fleet_ids):
        if fid not in available_troopbase_fleet_ids:  # entry may have been discarded in previous loop iterations
            continue
        fleet = universe.getFleet(fid)
        if not fleet:
            continue
        sys_id = fleet.systemID
        system = universe.getSystem(sys_id)
        available_planets = set(system.planetIDs).intersection(set(aistate.qualifyingTroopBaseTargets.keys()))
        debug(
            "Considering Base Troopers in {}, found planets {} and registered targets {} with status {}".format(
                system.name,
                list(system.planetIDs),
                available_planets,
                [(pid, aistate.qualifyingTroopBaseTargets[pid]) for pid in available_planets],
            )
        )
        targets = [pid for pid in available_planets if aistate.qualifyingTroopBaseTargets[pid][1] != -1]
        if not targets:
            debug("Failure: found no valid target for troop base in system %s" % system)
            continue
        status = aistate.systemStatus.get(sys_id, {})
        local_base_troops = set(status.get("myfleets", [])).intersection(available_troopbase_fleet_ids)

        target_id = INVALID_ID
        best_score = -1
        target_troops = 0
        for pid, (p_score, p_troops) in assign_invasion_values(targets).items():
            if p_score > best_score:
                best_score = p_score
                target_id = pid
                target_troops = p_troops
        if target_id == INVALID_ID:
            continue
        local_base_troops.discard(fid)
        found_fleets = []
        troops_needed = max(0, target_troops - FleetUtilsAI.count_troops_in_fleet(fid))
        found_stats = {}
        min_stats = {"troopCapacity": troops_needed}
        target_stats = {"troopCapacity": troops_needed + _TROOPS_SAFETY_MARGIN}

        FleetUtilsAI.get_fleets_for_mission(
            target_stats,
            min_stats,
            found_stats,
            starting_system=sys_id,
            fleet_pool_set=local_base_troops,
            fleet_list=found_fleets,
        )
        for fid2 in found_fleets:
            FleetUtilsAI.merge_fleet_a_into_b(fid2, fid)
            available_troopbase_fleet_ids.discard(fid2)
        available_troopbase_fleet_ids.discard(fid)
        aistate.qualifyingTroopBaseTargets[target_id][1] = -1  # TODO: should probably delete
        target = TargetPlanet(target_id)
        fleet_mission = aistate.get_fleet_mission(fid)
        fleet_mission.set_target(MissionType.ORBITAL_INVASION, target)


def assign_invasion_fleets_to_invade():
    """Assign fleet targets to invadable planets."""
    aistate = get_aistate()

    assign_invasion_bases()

    all_invasion_fleet_ids = FleetUtilsAI.get_empire_fleet_ids_by_role(MissionType.INVASION)
    invasion_fleet_ids = FleetUtilsAI.extract_fleet_ids_without_mission_types(all_invasion_fleet_ids)
    send_invasion_fleets(invasion_fleet_ids, AIstate.invasionTargets, MissionType.INVASION)
    all_invasion_fleet_ids = FleetUtilsAI.get_empire_fleet_ids_by_role(MissionType.INVASION)
    for fid in FleetUtilsAI.extract_fleet_ids_without_mission_types(all_invasion_fleet_ids):
        this_mission = aistate.get_fleet_mission(fid)
        this_mission.check_mergers(context="Post-send consolidation of unassigned troops")


def _get_base_troopers(sys_id: SystemId) -> float:
    """Get troopers already in system, including the bases in the build queue."""
    aistate = get_aistate()
    this_sys_status = aistate.systemStatus.get(sys_id, {})
    tally = 0
    for _fid in this_sys_status.get("myfleets", []):
        tally += FleetUtilsAI.count_troops_in_fleet(_fid)
    for element in fo.getEmpire().productionQueue:
        tally += _get_queued_base_troopers(sys_id, element)
    return tally


def _get_queued_base_troopers(sys_id: SystemId, element: fo.productionQueueElement) -> float:
    """Get base troopers for given system from a build queue element"""
    aistate = get_aistate()
    if element.buildType == EmpireProductionTypes.BT_SHIP:
        planet = fo.getUniverse().getPlanet(element.locationID)
        if planet.systemID == sys_id and aistate.get_ship_role(element.designID) == ShipRoleType.BASE_INVASION:
            design = fo.getShipDesign(element.designID)
            troops_per_ship = design.troopCapacity * get_species_attack_troops(planet.speciesName)
            return element.remaining * element.blocksize * troops_per_ship
    return 0.0



from collections.abc import Iterable
from functools import reduce
from itertools import chain
from typing import Union


def _combine_ratings(rating1: float, rating2: float) -> float:
    """Combines two combat ratings to a total rating.

    The formula takes into account the fact that the combined strength of two ships is more than the
    sum of its individual ratings. Basic idea as follows:

    We use the following definitions

    r: rating
    a: attack
    s: structure

    where we take into account effective values after accounting for e.g. shields effects.

    We generally define the rating of a ship as
    r_i = a_i*s_i                                                                   (1)

    A natural extension for the combined rating of two ships is
    r_tot = (a_1+a_2)*(s_1+s_2)                                                     (2)

    Assuming         a_i approx s_i                                                 (3)
    It follows that  a_i approx sqrt(r_i) approx s_i                                (4)
    And thus         r_tot = (sqrt(r_1)+sqrt(r_2))^2 = r1 + r2 + 2*sqrt(r1*r2)      (5)

    Note that this function has commutative and associative properties.
    """
    return rating1 + rating2 + 2 * (rating1 * rating2) ** 0.5


def combine_ratings(first: Union[float, Iterable], *ratings: float) -> float:
    """
    Combine multiple ratings to one.
    """
    if isinstance(first, (int, float)):
        first = (first,)
    return reduce(_combine_ratings, chain(first, ratings), 0.0)



from focs._tech import *
from macros.base_prod import TECH_COST_MULTIPLIER

Tech(
    name="CON_ASYMP_MATS",
    description="CON_ASYMP_MATS_DESC",
    short_description="THEORY_SHORT_DESC",
    category="CONSTRUCTION_CATEGORY",
    researchcost=20 * TECH_COST_MULTIPLIER,
    researchturns=2,
    tags=["PEDIA_CONSTRUCTION_CATEGORY", "THEORY"],
    graphic="icons/tech/asymptotic_materials.png",
)



from focs._tech import *
from macros.base_prod import TECH_COST_MULTIPLIER
from techs.ship_weapons.ship_weapons import WEAPON_UPGRADE_CAPACITY_EFFECTS

Tech(
    name="SHP_WEAPON_4_2",
    description="SHP_WEAPON_4_2_DESC",
    short_description="SHIP_WEAPON_IMPROVE_SHORT_DESC",
    category="SHIP_WEAPONS_CATEGORY",
    researchcost=600 * TECH_COST_MULTIPLIER,
    researchturns=3,
    tags=["PEDIA_SR_WEAPON_TECHS"],
    prerequisites=["SHP_WEAPON_4_1"],
    effectsgroups=WEAPON_UPGRADE_CAPACITY_EFFECTS("SHP_WEAPON_4_2", "SR_WEAPON_4_1", 5),
    graphic="icons/ship_parts/death-ray-2.png",
)



from focs._species import *
from species.species_macros.empire_opinions import COMMON_OPINION_EFFECTS
from species.species_macros.env import SWAMP_STANDARD_EP
from species.species_macros.focus import (
    HAS_ADVANCED_FOCI,
    HAS_GROWTH_FOCUS,
    HAS_INDUSTRY_FOCUS,
    HAS_INFLUENCE_FOCUS,
    HAS_RESEARCH_FOCUS,
)
from species.species_macros.happiness import AVERAGE_HAPPINESS
from species.species_macros.industry import AVERAGE_INDUSTRY
from species.species_macros.influence import GREAT_INFLUENCE
from species.species_macros.native_fortification import DEFAULT_NATIVE_DEFENSE
from species.species_macros.population import GOOD_POPULATION
from species.species_macros.research import AVERAGE_RESEARCH
from species.species_macros.shields import STANDARD_SHIP_SHIELDS
from species.species_macros.stockpile import AVERAGE_STOCKPILE
from species.species_macros.supply import AVERAGE_SUPPLY
from species.species_macros.troops import BAD_DEFENSE_TROOPS, BAD_OFFENSE_TROOPS
from species.species_macros.weapons import BAD_WEAPONS

Species(
    name="SP_TAEGHIRUS",
    description="SP_TAEGHIRUS_DESC",
    gameplay_description="SP_TAEGHIRUS_GAMEPLAY_DESC",
    native=True,
    can_produce_ships=True,
    can_colonize=True,
    tags=[
        "ORGANIC",
        "TELEPATHIC",
        "GOOD_POPULATION",
        "GREAT_INFLUENCE",
        "BAD_WEAPONS",
        "AVERAGE_SUPPLY",
        "BAD_OFFENSE_TROOPS",
        "PEDIA_ORGANIC_SPECIES_CLASS",
        "PEDIA_TELEPATHIC_TITLE",
    ],
    foci=[
        HAS_INDUSTRY_FOCUS,
        HAS_RESEARCH_FOCUS,
        HAS_INFLUENCE_FOCUS,
        HAS_GROWTH_FOCUS,
        *HAS_ADVANCED_FOCI,
    ],
    defaultfocus="FOCUS_INFLUENCE",
    likes=[
        "FOCUS_INFLUENCE",
        "SHIMMER_SILK_SPECIAL",
        "SPARK_FOSSILS_SPECIAL",
        "WORLDTREE_SPECIAL",
        "MINERALS_SPECIAL",
        "FRUIT_SPECIAL",
        "HONEYCOMB_SPECIAL",
        "PLC_CONFORMANCE",
        "PLC_MODERATION",
        "PLC_POPULATION",
        "PLC_CENTRALIZATION",
        "PLC_INDOCTRINATION",
    ],
    dislikes=[
        "BLD_INTERSPECIES_ACADEMY",
        "BLD_CULTURE_ARCHIVES",
        "BLD_CULTURE_LIBRARY",
        "BLD_COLLECTIVE_NET",
        "BLD_PLANET_BEACON",
        "BLD_GENOME_BANK",
        "BLD_COLLECTIVE_NET",
        "RESONANT_MOON_SPECIAL",
        "TIDAL_LOCK_SPECIAL",
        "ECCENTRIC_ORBIT_SPECIAL",
        "KRAKEN_NEST_SPECIAL",
        "PLC_MARINE_RECRUITMENT",
        "PLC_TERROR_SUPPRESSION",
        "PLC_MARTIAL_LAW",
        "PLC_DIVINE_AUTHORITY",
        "SP_NIGHTSIDERS",
    ],
    effectsgroups=[
        *AVERAGE_INDUSTRY,
        *AVERAGE_RESEARCH,
        *GREAT_INFLUENCE,
        *AVERAGE_STOCKPILE,
        *GOOD_POPULATION,
        *AVERAGE_HAPPINESS,
        COMMON_OPINION_EFFECTS("SP_TAEGHIRUS"),
        *AVERAGE_SUPPLY,
        *BAD_DEFENSE_TROOPS,
        *BAD_OFFENSE_TROOPS,
        *BAD_WEAPONS,
        # not for description,
        *DEFAULT_NATIVE_DEFENSE,
        *STANDARD_SHIP_SHIELDS,
    ],
    environments=SWAMP_STANDARD_EP,
    graphic="icons/species/t-aeghirus.png",
)



from focs._tech import *
from macros.base_prod import TECH_COST_MULTIPLIER

Tech(
    name="CON_ART_PLANET",
    description="CON_ART_PLANET_DESC",
    short_description="BUILDING_UNLOCK_SHORT_DESC",
    category="CONSTRUCTION_CATEGORY",
    researchcost=440 * TECH_COST_MULTIPLIER,
    researchturns=8,
    tags=["PEDIA_CONSTRUCTION_CATEGORY"],
    prerequisites=["CON_ART_HEAVENLY", "LRN_SPATIAL_DISTORT_GEN"],
    unlock=[
        Item(type=UnlockBuilding, name="BLD_ART_PLANET"),
        Item(type=UnlockBuilding, name="BLD_ART_FACTORY_PLANET"),
        Item(type=UnlockBuilding, name="BLD_ART_PARADISE_PLANET"),
    ],
    graphic="icons/tech/artificial_planet.png",
)



from focs._species import *
from species.species_macros.empire_opinions import COMMON_OPINION_EFFECTS
from species.species_macros.env import SWAMP_STANDARD_EP
from species.species_macros.focus import (
    HAS_ADVANCED_FOCI,
    HAS_GROWTH_FOCUS,
    HAS_INDUSTRY_FOCUS,
    HAS_INFLUENCE_FOCUS,
    HAS_RESEARCH_FOCUS,
)
from species.species_macros.happiness import AVERAGE_HAPPINESS
from species.species_macros.industry import GOOD_INDUSTRY
from species.species_macros.influence import BAD_INFLUENCE
from species.species_macros.planet_defense import AVERAGE_PLANETARY_DEFENSE
from species.species_macros.planet_shields import AVERAGE_PLANETARY_SHIELDS
from species.species_macros.planet_size import LARGE_PLANET
from species.species_macros.population import AVERAGE_POPULATION
from species.species_macros.research import GOOD_RESEARCH
from species.species_macros.shields import STANDARD_SHIP_SHIELDS
from species.species_macros.stockpile import AVERAGE_STOCKPILE
from species.species_macros.supply import AVERAGE_SUPPLY
from species.species_macros.troops import BAD_DEFENSE_TROOPS, BAD_OFFENSE_TROOPS
from species.species_macros.weapons import BAD_WEAPONS

Species(
    name="SP_GYSACHE",
    description="SP_GYSACHE_DESC",
    gameplay_description="SP_GYSACHE_GAMEPLAY_DESC",
    playable=True,
    can_produce_ships=True,
    can_colonize=True,
    tags=[
        "AVERAGE_POPULATION",
        "ORGANIC",
        "GOOD_INDUSTRY",
        "GOOD_RESEARCH",
        "BAD_INFLUENCE",
        "BAD_WEAPONS",
        "AVERAGE_SUPPLY",
        "BAD_OFFENSE_TROOPS",
        "PEDIA_ORGANIC_SPECIES_CLASS",
    ],
    foci=[
        HAS_INDUSTRY_FOCUS,
        HAS_RESEARCH_FOCUS,
        HAS_INFLUENCE_FOCUS,
        HAS_GROWTH_FOCUS,
        *HAS_ADVANCED_FOCI,
    ],
    defaultfocus="FOCUS_INDUSTRY",
    likes=[
        "FOCUS_INDUSTRY",
        "FRACTAL_GEODES_SPECIAL",
        "SUCCULENT_BARNACLES_SPECIAL",
        "GAIA_SPECIAL",
        "HONEYCOMB_SPECIAL",
        "ELERIUM_SPECIAL",
        "CRYSTALS_SPECIAL",
        "PLC_BOOTSTRAPPING",
        "PLC_CHECKPOINTS",
        "PLC_COLONIZATION",
        "PLC_CONFORMANCE",
        "PLC_CONTINUOUS_SCANNING",
        "PLC_INSURGENCY",
    ],
    dislikes=[
        "BLD_SPACE_ELEVATOR",
        "BLD_NEUTRONIUM_SYNTH",
        "BLD_NEUTRONIUM_FORGE",
        "BLD_NEUTRONIUM_EXTRACTOR",
        "BLD_SCRYING_SPHERE",
        "BLD_SHIPYARD_ENRG_COMP",
        "RESONANT_MOON_SPECIAL",
        "TIDAL_LOCK_SPECIAL",
        "WORLDTREE_SPECIAL",
        "MONOPOLE_SPECIAL",
        "PLC_CENTRALIZATION",
        "PLC_EXPLORATION",
        "PLC_ISOLATION",
        "PLC_MARINE_RECRUITMENT",
        "PLC_MARTIAL_LAW",
        "PLC_THE_HUNT",
        "SP_NIGHTSIDERS",
    ],
    effectsgroups=[
        *GOOD_INDUSTRY,
        *GOOD_RESEARCH,
        *BAD_INFLUENCE,
        *AVERAGE_STOCKPILE,
        *AVERAGE_POPULATION,
        *AVERAGE_HAPPINESS,
        COMMON_OPINION_EFFECTS("SP_GYSACHE"),
        *AVERAGE_SUPPLY,
        *BAD_DEFENSE_TROOPS,
        *BAD_OFFENSE_TROOPS,
        *BAD_WEAPONS,
        # not for description,
        *AVERAGE_PLANETARY_SHIELDS,
        *AVERAGE_PLANETARY_DEFENSE,
        *LARGE_PLANET,
        *STANDARD_SHIP_SHIELDS,
    ],
    environments=SWAMP_STANDARD_EP,
    graphic="icons/species/gysache.png",
)
